import numpy as np
from scipy.interpolate import RegularGridInterpolator

from lasy.utils.exp_data_utils import find_center_of_mass

from .transverse_profile import TransverseProfile


class TransverseProfileFromData(TransverseProfile):
    """
    Class for transverse laser profile.

    Created using data from an experimental measurement or from
    the output of another code.

    Uses user supplied data to define the transverse profile
    of the laser pulse.

    The data must be supplied as a 2D numpy array of intensity
    values (for example an imported camera image from an
    experimental measurement).

    In the case of experimental measurements, this data
    should already have some undergone some preprocessing
    such as background subtraction and noise removal.

    The beam will be imported and automatically centered unless
    otherwise specified.

    Parameters
    ----------
    intensity_data : 2Darray of floats
        The 2D transverse intensity profile of the laser pulse.

    lo, hi : list of scalars (in meters)
        Lower and higher end of the physical domain of the data.
        One element per direction (in this case 2)

    center_data : bool, optional
        If true, the intensity data will be rolled to put the
        center of mass at the center of the image. It will
        also shift the x and y data axes such that (x,y) = (0,0)
        is also located at the center of the image. Default is True
    """

    def __init__(self, intensity_data, lo, hi, center_data=True):
        super().__init__()

        intensity_data = intensity_data.astype("float64")

        n_y, n_x = np.shape(intensity_data)

        dx = (hi[0] - lo[0]) / n_x
        dy = (hi[1] - lo[1]) / n_y

        if center_data:
            x_range = np.abs(hi[0] - lo[0])
            y_range = np.abs(hi[1] - lo[1])
            x_data = np.linspace(-x_range / 2, x_range / 2, n_x)
            y_data = np.linspace(-y_range / 2, y_range / 2, n_y)

            n_x0, n_y0 = find_center_of_mass(intensity_data)
            intensity_data = np.roll(
                np.roll(intensity_data, -int(n_x0 - n_x / 2), axis=1),
                -int(n_y0 - n_y / 2),
                axis=0,
            )

        else:
            x_data = np.linspace(lo[0], hi[0], n_x)
            y_data = np.linspace(lo[1], hi[1], n_y)

        # Normalise the profile such that its squared integeral == 1
        intensity_data /= np.sum(intensity_data) * dx * dy

        # Note here we use the square root of intensity to get the 'field'
        self.field_interp = RegularGridInterpolator(
            (y_data, x_data),
            np.sqrt(intensity_data),
            bounds_error=False,
            fill_value=0.0,
        )

    def _evaluate(self, x, y):
        """
        Return the transverse envelope.

        Parameters
        ----------
        x, y : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of floats
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y
        """
        envelope = self.field_interp((y, x))

        return envelope



import numpy as np

from .longitudinal_profile import LongitudinalProfile


class ContinuousWaveProfile(LongitudinalProfile):
    r"""
    Class representing a continous wave laser longitudinal profile.

    Specifically, the longitudinal profile will be represented by a constant value.

    Note that using this profile will force the number of points in the longitudinal
    dimension to 1. Additionally, it will force the length of the longitudinal axis
    1 s.

    Parameters
    ----------
    wavelength : float (in meter)
        The main laser wavelength :math:`\lambda_0` of the laser.
    """

    def __init__(self, wavelength):
        super().__init__(wavelength)
        self.__update_is_cw__(True)

    def evaluate(self, t):
        """
        Return the longitudinal envelope.

        Parameters
        ----------
        t : ndarrays of floats
            Define longitudinal points on which to evaluate the envelope

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the longitudinal envelope at the
            specified points. This array has the same shape as the array t.
        """
        return np.ones_like(t + 0j)



# -*- coding: utf-8 -*-
"""Thest the parabolic mirror implementation.

Test checks the implementation of the parabolic mirror
by initializing a Gaussian pulse in the near field, and
propagating it through a parabolic mirror, and then to
the focal position ; we then check that the waist as the
expected value in the far field (i.e. in the focal plane).
"""

import numpy as np

from lasy.laser import Laser
from lasy.optical_elements import ParabolicMirror
from lasy.profiles.gaussian_profile import GaussianProfile

wavelength = 0.8e-6
w0 = 5.0e-3  # m, initialized in near field

# The laser is initialized in the near field
pol = (1, 0)
laser_energy = 1.0  # J
t_peak = 0.0e-15  # s
tau = 30.0e-15  # s
gaussian_profile = GaussianProfile(wavelength, pol, laser_energy, w0, tau, t_peak)


def get_w0(laser):
    # Calculate the laser waist
    field = laser.grid.get_temporal_field()
    if laser.dim == "xyt":
        Nx = field.shape[0]
        A2 = (np.abs(field[Nx // 2 - 1, :, :]) ** 2).sum(-1)
        ax = laser.grid.axes[1]
    else:
        A2 = (np.abs(field[0, :, :]) ** 2).sum(-1)
        ax = laser.grid.axes[0]
        if ax[0] > 0:
            A2 = np.r_[A2[::-1], A2]
            ax = np.r_[-ax[::-1], ax]
        else:
            A2 = np.r_[A2[::-1][:-1], A2]
            ax = np.r_[-ax[::-1][:-1], ax]

    sigma = 2 * np.sqrt(np.average(ax**2, weights=A2))

    return sigma


def check_parabolic_mirror(laser):
    # Propagate laser after parabolic mirror + vacuum
    f0 = 8.0  # focal distance in m
    laser.apply_optics(ParabolicMirror(f=f0))
    laser.propagate(f0)
    # Check that the value is the expected one in the near field
    w0_num = get_w0(laser)
    w0_theor = wavelength * f0 / (np.pi * w0)
    err = 2 * np.abs(w0_theor - w0_num) / (w0_theor + w0_num)
    assert err < 1e-3


def test_3D_case():
    # - 3D case
    # The laser is initialized in the near field
    dim = "xyt"
    lo = (-12e-3, -12e-3, -60e-15)
    hi = (+12e-3, +12e-3, +60e-15)
    npoints = (500, 500, 100)

    laser = Laser(dim, lo, hi, npoints, gaussian_profile)
    check_parabolic_mirror(laser)


def test_RT_case():
    # - Cylindrical case
    # The laser is initialized in the near field
    dim = "rt"
    lo = (0e-6, -60e-15)
    hi = (15e-3, +60e-15)
    npoints = (750, 100)

    laser = Laser(dim, lo, hi, npoints, gaussian_profile)
    check_parabolic_mirror(laser)



import numpy as np

from .longitudinal_profile import LongitudinalProfile


class GaussianLongitudinalProfile(LongitudinalProfile):
    r"""
    Class for the analytic profile of a longitudinally-Gaussian laser pulse.

    More precisely, the longitudinal envelope
    (to be used in the :class:`.CombinedLongitudinalTransverseProfile` class)
    corresponds to:

    .. math::

        \mathcal{L}(t) = \exp\left( - \frac{(t-t_{peak})^2}{\tau^2}
                        + i\omega_0t_{peak} \right)

    Parameters
    ----------
    tau : float (in second)
        The duration of the laser pulse, i.e. :math:`\tau` in the above
        formula. Note that :math:`\tau = \tau_{FWHM}/\sqrt{2\log(2)}`,
        where :math:`\tau_{FWHM}` is the Full-Width-Half-Maximum duration
        of the intensity distribution of the pulse.

    t_peak : float (in second)
        The time at which the laser envelope reaches its maximum amplitude,
        i.e. :math:`t_{peak}` in the above formula.

    cep_phase : float (in radian), optional
        The Carrier Enveloppe Phase (CEP), i.e. :math:`\phi_{cep}`
        in the above formula (i.e. the phase of the laser
        oscillation, at the time where the laser envelope is maximum).
    """

    def __init__(self, wavelength, tau, t_peak, cep_phase=0):
        super().__init__(wavelength)
        self.tau = tau
        self.t_peak = t_peak
        self.cep_phase = cep_phase

    def evaluate(self, t):
        """
        Return the longitudinal envelope.

        Parameters
        ----------
        t : ndarrays of floats
            Define longitudinal points on which to evaluate the envelope

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the longitudinal envelope at the
            specified points. This array has the same shape as the array t.
        """
        envelope = np.exp(
            -((t - self.t_peak) ** 2) / self.tau**2
            + 1.0j * (self.cep_phase + self.omega0 * self.t_peak)
        )
        return envelope



import numpy as np
from scipy.constants import c, pi


class LongitudinalProfile(object):
    """
    Base class for longitudinal profiles of laser pulses.

    Any new longitudinal profile should inherit from this class, and define
    its own `evaluate` method, using the same signature as the method below.
    """

    def __init__(self, wavelength):
        self.lambda0 = wavelength
        self.omega0 = 2 * pi * c / self.lambda0
        self.is_cw = False

    def evaluate(self, t):
        """
        Return the longitudinal envelope.

        Parameters
        ----------
        t : ndarrays of floats
            Define points on which to evaluate the envelope

        Returns
        -------
        envelope: ndarray of complex numbers
            Contains the value of the longitudinal envelope at the
            specified points. This array has the same shape as the array t.
        """
        # The base class only defines dummy fields
        # (This should be replaced by any class that inherits from this one.)
        return np.zeros(t.shape, dtype="complex128")

    def __update_is_cw__(self, value):
        """Update state of is_cw variable."""
        self.is_cw = value



"""Test the implementation of the HG recostruction.

Test checks the implementation of the HG reconstruction. It does so
by initializing a super-Gaussian pulse and denoising it. It then
checks that the error remains positive and less than the predefined
value.
"""

import numpy as np

from lasy.profiles.transverse.super_gaussian_profile import (
    SuperGaussianTransverseProfile,
)
from lasy.utils.denoise import hg_reconstruction


def test_denoise_hg_reconstruction():
    # Parameters
    w0 = 20e-6
    shape_parameter = 3
    wavelength = 8e-7
    resolution = 0.2e-6
    lo = [-2e-4, -2e-4]
    hi = [2e-4, 2e-4]

    # Define the transverse profile
    transverse_profile = SuperGaussianTransverseProfile(
        w0, shape_parameter
    )  # Super-Gaussian profile
    transverse_profile_cleaned, w0x, w0y = hg_reconstruction(
        transverse_profile, wavelength, resolution, lo, hi
    )  # Denoised profile

    # Calculate the error
    x = np.linspace(-5 * w0x, 5 * w0x, 500)
    X, Y = np.meshgrid(x, x)
    prof1 = transverse_profile.evaluate(X, Y)
    prof2 = transverse_profile_cleaned.evaluate(X, Y)
    error = np.sum(np.abs(prof2 - prof1) ** 2) / np.sum(np.abs(prof1) ** 2)
    assert error < 0.02



# -*- coding: utf-8 -*-
"""Test the polynomial spectral phase implementation.

Test checks the implementation of the polynomial spectral phase
by initializing a Gaussian pulse (with flat spectral phase),
adding spectral phase to it, and checking the corresponding
temporal shape of the laser pulse against analytical formulas.
"""

import numpy as np
from scipy.constants import c

from lasy.laser import Laser
from lasy.optical_elements import PolynomialSpectralPhase
from lasy.profiles.gaussian_profile import GaussianProfile

# Laser parameters
wavelength = 0.8e-6
omega0 = 2 * np.pi * c / wavelength
w0 = 5.0e-6  # m
pol = (1, 0)
laser_energy = 1.0  # J
t_peak = 0.0e-15  # s
tau = 15.0e-15  # s
gaussian_profile = GaussianProfile(wavelength, pol, laser_energy, w0, tau, t_peak)

# Grid parameters
dim = "xyt"
lo = (-12e-6, -12e-6, -100e-15)
hi = (+12e-6, +12e-6, +100e-15)
npoints = (100, 100, 200)


def test_delay():
    """Add and compare group delay for laser.

    Add group delay to the laser and compare the on-axis field with the
    analytical formula for a Gaussian pulse with group delay.
    """
    delay = 50e-15
    dazzler = PolynomialSpectralPhase(omega0, delay=delay)

    # Initialize the laser
    dim = "xyt"
    lo = (-12e-6, -12e-6, -100e-15)
    hi = (+12e-6, +12e-6, +100e-15)
    npoints = (100, 100, 400)
    laser = Laser(dim, lo, hi, npoints, gaussian_profile)

    # Get field before and after dazzler
    E_before = laser.grid.get_temporal_field()
    laser.apply_optics(dazzler)
    E_after = laser.grid.get_temporal_field()

    # Extract peak field before dazzler
    E0 = abs(E_before[50, 50]).max()

    # Compute the analtical expression in real space for a Gaussian
    t = np.linspace(laser.grid.lo[-1], laser.grid.hi[-1], laser.grid.npoints[-1])

    E_analytical = E0 * np.exp(-1.0 * ((t - delay) / tau) ** 2)

    # Compare the on-axis field with the analytical formula
    tol = 1.2e-3
    assert np.all(
        abs(E_after[50, 50, :] - E_analytical) / abs(E_analytical).max() < tol
    )


def test_gdd():
    """Add and compare GDD laser.

    Add GDD to the laser and compare the on-axis field with the
    analytical formula for a Gaussian pulse with GDD.
    """
    gdd = 200e-30
    dazzler = PolynomialSpectralPhase(omega0, gdd=gdd)

    # Initialize the laser
    dim = "xyt"
    lo = (-12e-6, -12e-6, -100e-15)
    hi = (+12e-6, +12e-6, +100e-15)
    npoints = (100, 100, 200)
    laser = Laser(dim, lo, hi, npoints, gaussian_profile)

    # Get field before and after dazzler
    E_before = laser.grid.get_temporal_field()
    laser.apply_optics(dazzler)
    E_after = laser.grid.get_temporal_field()
    # Extract peak field before dazzler
    E0 = abs(E_before[50, 50]).max()

    # Compute the analtical expression in real space for a Gaussian
    t = np.linspace(laser.grid.lo[-1], laser.grid.hi[-1], laser.grid.npoints[-1])
    stretch_factor = 1 - 2j * gdd / tau**2
    E_analytical = (
        E0 * np.exp(-1.0 / stretch_factor * (t / tau) ** 2) / stretch_factor**0.5
    )

    # Compare the on-axis field with the analytical formula
    tol = 1.2e-3
    assert np.all(
        abs(E_after[50, 50, :] - E_analytical) / abs(E_analytical).max() < tol
    )


def test_tod():
    """Add and compare TOD laser.

    Add TOD to the laser and compare the on-axis field with the
    analytical formula from the stationary phase approximation.
    """
    tod = 5e-42
    dazzler = PolynomialSpectralPhase(omega0, tod=tod)

    # Initialize the laser
    dim = "xyt"
    lo = (-12e-6, -12e-6, -50e-15)
    hi = (+12e-6, +12e-6, +250e-15)
    npoints = (100, 100, 400)
    laser = Laser(dim, lo, hi, npoints, gaussian_profile)
    t = np.linspace(laser.grid.lo[-1], laser.grid.hi[-1], laser.grid.npoints[-1])

    # Get field before and after dazzler
    E_before = laser.grid.get_temporal_field()
    laser.apply_optics(dazzler)
    E_after = laser.grid.get_temporal_field()
    # Extract peak field before dazzler
    E0 = abs(E_before[50, 50]).max()

    # Compare data in the post-pulse region to the stationary phase approximation.
    # The stationary phase approximation result was obtained under the additional
    # assumption t >> tau^4/tod, so we only compare the data in this region
    E_compare = abs(E_after[50, 50, t > t_peak + 2 * tau**4 / tod])  # On-axis field
    t = t[t > t_peak + 2 * tau**4 / tod]
    prediction = abs(
        2
        * E0
        * tau
        / (8 * tod * t) ** 0.25
        * np.exp(-(tau**2) * t / (2 * tod))
        * np.cos(
            2 * t / 3 * (2 * t / tod) ** 0.5
            - tau**4 / (8 * tod) * (2 * t / tod) ** 0.5
            - np.pi / 4
        )
    )

    # Compare the on-axis field with the analytical formula
    tol = 2.4e-2
    assert np.all(abs(E_compare - prediction) / abs(E0) < tol)



from .continuous_wave_profile import ContinuousWaveProfile
from .cosine_profile import CosineLongitudinalProfile
from .gaussian_profile import GaussianLongitudinalProfile
from .longitudinal_profile import LongitudinalProfile
from .longitudinal_profile_from_data import LongitudinalProfileFromData
from .super_gaussian_profile import SuperGaussianLongitudinalProfile

__all__ = [
    "CosineLongitudinalProfile",
    "GaussianLongitudinalProfile",
    "SuperGaussianLongitudinalProfile",
    "LongitudinalProfileFromData",
    "LongitudinalProfile",
    "ContinuousWaveProfile",
]



from math import factorial

import numpy as np
from scipy.special import genlaguerre

from .transverse_profile import TransverseProfile


class LaguerreGaussianTransverseProfile(TransverseProfile):
    r"""
    A high-order Gaussian laser pulse expressed in the Laguerre-Gaussian formalism.

    Definition is according to Siegman "Lasers" pg. 646 eq. 64.

    More precisely, the transverse envelope (to be used in the
    :class:`.CombinedLongitudinalTransverseLaser` class) corresponds to:

    .. math::
        \mathcal{T}(x, y) = \,
                \mathcal{L}_{p,m} (x) \, \exp(i \Phi)

    with

    .. math::
        \mathcal{L}_{p,m}(x) = A \left ( \frac{\sqrt{2}r}{w(z)} \right)^m l_{p,m} \left ( \frac{2 r^2}{w^2(z)} \right)
        \exp{\left( -\frac{r^2}{w^2(z)}\right)} \exp{ \left ( -i k_0 \frac{r^2}{2 R(z)} \right )}

        w(z) = w_{0} \sqrt{1 + \left( \frac{z}{Z_R}\right)^2}

        A = \frac{1}{w(z)} \sqrt{\frac{2 p!}{\pi (p + m)!}}

        R(z) = z + \frac{Z_R^2}{z}

        \Phi(z) = \left(2 p + m + 1\right) \arctan\left({\frac{z}{Z_R}}\right)

        Z_R = \frac{\pi w_0^2}{\lambda_0}


    where  :math:`l_{p,m}` is the Laguerre polynomial of radial order :math:`p` and azimuthal order :math:`m`.

    The z-depedence shown in the above equations is required to correctly define the
    electric field of the transverse profile relative to that of the pulse at the focus.
    The absolute z position will be overwritten when creating a laser object.

    Parameters
    ----------
    w_0 : float (in meter)
        The waist of the laser pulse,
        i.e. :math:`w_{0}` in the above formula.
    p : int (dimensionless)
        The order of Laguerre polynomial in the x direction
        i.e. :math:`m` in the above formula.
    m : int (dimensionless)
        The order of Laguerre polynomial in the y direction
        i.e. :math:`n` in the above formula.
    wavelength : float (in meter)
        The main laser wavelength :math:`\lambda_0` of the laser.
    z_foc : float (in meter), optional
        Position of the focal plane. (The laser pulse is initialized at
        ``z=0``.)

    Warnings
    --------
    In order to initialize the pulse out of focus, you can either:

    - Use a non-zero ``z_foc``
    - Use ``z_foc=0`` (i.e. initialize the pulse at focus) and then call
      ``laser.propagate(-z_foc)``

    Both methods are in principle equivalent, but note that the first
    method uses the paraxial approximation, while the second method does
    not make this approximation.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> from lasy.profiles.transverse.laguerre_gaussian_profile import (
    ...     LaguerreGaussianTransverseProfile,
    ... )
    >>> # Create evaluation grid
    >>> xy = np.linspace(-30e-6, 30e-6, 200)
    >>> X, Y = np.meshgrid(xy, xy)
    >>> # Create an array of plots
    >>> fig, ax = plt.subplots(3, 6, figsize=(10, 5), tight_layout=True)
    >>> extent = (1e6 * xy[0], 1e6 * xy[-1], 1e6 * xy[0], 1e6 * xy[-1])
    >>> for p in range(3):
    >>>     for m in range(3):
    >>>         transverse_profile = LaguerreGaussianTransverseProfile(
    ...             w_0 = 10e-6, # m
    ...             p = p, #
    ...             m = m, #
    ...             wavelength = 0.8e-6, # m
    ...         )
    ...         intensity = np.abs(transverse_profile.evaluate(X,Y))**2
    ...         vmax_intensity = np.max(intensity)
    >>>         ax[p,m].imshow(intensity,extent=extent,cmap='bone_r',vmin=0,vmax=vmax_intensity)
    >>>         ax[p,m].set_title('Inten: p,m = %i,%i' %(p,m))
    >>>         phase = np.angle(transverse_profile.evaluate(X,Y))
    ...         vmax_phase = np.max(np.abs(phase))
    >>>         ax[p,m+3].imshow(phase,extent=extent,cmap='seismic',vmin=-vmax_phase,vmax=vmax_phase)
    >>>         ax[p,m+3].set_title('Phase: p,m = %i,%i' %(p,m))
    >>>         if p==2:
    >>>             ax[p,m].set_xlabel("x (µm)")
    >>>             ax[p,m+3].set_xlabel("x (µm)")
    >>>         else:
    >>>             ax[p,m].set_xticks([])
    >>>             ax[p,m+3].set_xticks([])
    >>>         if m==0:
    >>>             ax[p,m].set_ylabel("y (µm)")
    >>>             ax[p,m+3].set_yticks([])
    >>>         else:
    >>>             ax[p,m].set_yticks([])
    >>>             ax[p,m+3].set_yticks([])



    """

    def __init__(self, w_0, p, m, wavelength, z_foc=0):
        super().__init__()
        self.w_0 = w_0
        self.p = p
        self.m = m
        self.wavelength = wavelength
        self.z_foc = z_foc
        z_eval = -z_foc  # this links our observation position to Siegman's definition

        self.k0 = 2 * np.pi / wavelength

        # Calculate Rayleigh Length
        Zr = np.pi * w_0**2 / wavelength

        # Calculate Size at Location Z
        w0Z = w_0 * np.sqrt(1 + (z_eval / Zr) ** 2)

        # Calculate Multiplicative Factors
        A = np.sqrt(2.0 * factorial(p) / (np.pi * factorial(m + p))) / w0Z

        # Calculate the Phase contributions from propagation
        phiZ = (2.0 * p + m + 1) * np.arctan2(z_eval, Zr)

        self.z_eval = z_eval
        self.Zr = Zr
        self.w0Z = w0Z
        self.A = A
        self.phiZ = phiZ

    def _evaluate(self, x, y):
        """
        Return the transverse envelope.

        Parameters
        ----------
        x, y : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y
        """
        z_eval = self.z_eval
        Zr = self.Zr
        k0 = self.k0
        w0Z = self.w0Z
        A = self.A
        p = self.p
        m = self.m
        phiZ = self.phiZ

        # Calculate the LG in each plane
        LG = (
            A
            * (np.sqrt(2.0) * np.sqrt(x**2 + y**2) / w0Z) ** (m)
            * genlaguerre(p, m)(2.0 * (x**2 + y**2) / w0Z**2)
            * np.exp(-(x**2 + y**2) / w0Z**2)
            * np.exp(-1j * k0 * (x**2 + y**2) / 2 / (z_eval**2 + Zr**2) * z_eval)
        )

        # Put it altogether
        envelope = LG * np.exp(1j * phiZ) * np.exp(-1j * m * np.arctan2(y, x))

        return envelope



"""Test the implementation of the axiparabola.

Test checks the implementation of the axiparabola
by initializing a super-Gaussian pulse in the near field, and
propagating it to the middle of the focal range. It then
checks that the field amplitude remains high over the focal range.
"""

import numpy as np

from lasy.laser import Laser
from lasy.optical_elements import Axiparabola
from lasy.profiles.combined_profile import CombinedLongitudinalTransverseProfile
from lasy.profiles.longitudinal import GaussianLongitudinalProfile
from lasy.profiles.transverse import SuperGaussianTransverseProfile


def test_axiparabola():
    # Define the laser profile
    wavelength = 800e-9  # Laser wavelength in meters
    polarization = (1, 0)  # Linearly polarized in the x direction
    energy = 1.5  # Energy of the laser pulse in joules
    spot_size = 1e-3  # Spot size in the near-field: millimeter-scale
    pulse_duration = 30e-15  # Pulse duration of the laser in seconds
    t_peak = 0.0  # Location of the peak of the laser pulse in time
    laser_profile = CombinedLongitudinalTransverseProfile(
        wavelength,
        polarization,
        GaussianLongitudinalProfile(wavelength, pulse_duration, t_peak),
        SuperGaussianTransverseProfile(spot_size, n_order=16),
        laser_energy=energy,
    )

    # Define the laser on a grid
    dimensions = "rt"  # Use cylindrical geometry
    lo = (0, -2.5 * pulse_duration)  # Lower bounds of the simulation box
    hi = (1.1 * spot_size, 2.5 * pulse_duration)  # Upper bounds of the simulation box
    num_points = (3000, 30)  # Number of points in each dimension
    laser = Laser(dimensions, lo, hi, num_points, laser_profile)

    # Define the parameters of the axiparabola
    f0 = 3e-2  # Focal distance
    delta = 1.5e-2  # Focal range
    R = spot_size  # Radius
    axiparabola = Axiparabola(f0, delta, R)

    # Apply axiparabola and propagate to the middle of the focal range
    laser.apply_optics(axiparabola)
    laser.propagate(f0 + delta / 2)
    E_middle = abs(laser.grid.get_temporal_field()).max()

    # Compute the equivalent Rayleigh length
    import math

    ZR = math.pi * wavelength * f0**2 / spot_size**2
    # Check that the focal range is much longer than the Rayleigh length
    assert delta > 5 * ZR

    laser.propagate(
        -2 * ZR
    )  # Propagate to two Rayleigh lengths before the middle of the focal range
    E_before = abs(laser.grid.get_temporal_field()).max()
    # For a Gaussian beam, the field amplitude should be reduced
    # by a factor of sqrt(5) after two Rayleigh lengths
    # Here, we check that the field remains significantly higher (1.5 times higher)
    assert E_before > E_middle / np.sqrt(5) * 1.5

    laser.propagate(
        4 * ZR
    )  # Propagate to two Rayleigh lengths after the middle of the focal range
    E_after = abs(laser.grid.get_temporal_field()).max()
    # For a Gaussian beam, the field amplitude should be reduced
    # by a factor of sqrt(5) after two Rayleigh lengths
    # Here, we check that the field remains significantly higher (1.5 times higher)
    assert E_after > E_after / np.sqrt(5) * 1.5



from lasy.profiles.transverse.hermite_gaussian_profile import (
    HermiteGaussianTransverseProfile,
)
from lasy.profiles.transverse.transverse_profile import TransverseProfile
from lasy.utils.mode_decomposition import hermite_gauss_decomposition


def hg_reconstruction(
    transverse_profile,
    wavelength,
    resolution,
    lo,
    hi,
    n_modes_x=10,
    n_modes_y=10,
):
    """
    Denoise the transverse profile by decomposing it into a set of Hermite-Gaussian modes.

    The profiles are weighted according to mode coefficients and then added.

    Parameters
    ----------
    transverse_profile : TransverseProfile object
        An instance of a class or sub-class of TransverseProfile.
        Defines the transverse envelope of the laser.

    wavelength : float (in meter)
        Central wavelength at which the Hermite-Gauss beams are to be defined.

    resolution : float
        The number of grid points in x and y used during the decomposition calculation.

    lo, hi : array of floats
        The lower and upper bounds of the spatial grid on which the
        decomposition is be performed.

    n_modes_x, n_modes_y : ints (optional)
        The maximum values of `n_x` and `n_y` up to which the expansion is performed.

    Returns
    -------
    transverse_profile_cleaned : class instance
        Denoised transverse profile after decomposition and recombination.

    w0x, w0y : floats
        Beam waist for which the decomposition is calculated.
        It is computed as the waist for which the weight of order 0 is maximum.
    """
    assert isinstance(transverse_profile, TransverseProfile)

    # Calculate the decomposition and waist of the laser pulse
    modeCoeffs, w0x, w0y = hermite_gauss_decomposition(
        transverse_profile, wavelength, resolution, lo, hi, n_modes_x, n_modes_y
    )

    # Denosing the laser profile
    for i, mode_key in enumerate(list(modeCoeffs)):
        transverse_profile_temp = HermiteGaussianTransverseProfile(
            w0x, w0y, mode_key[0], mode_key[1], wavelength
        )  # Create a new profile for each mode

        if i == 0:  # First mode (0,0)
            transverse_profile_cleaned = modeCoeffs[mode_key] * transverse_profile_temp
        else:  # All other modes
            transverse_profile_cleaned += modeCoeffs[mode_key] * transverse_profile_temp

    return transverse_profile_cleaned, w0x, w0y



import numpy as np


class TransverseProfile(object):
    """
    Base class for all transverse profiles.

    Any new transverse profile should inherit from this class, and define its own
    `evaluate` method, using the same signature as the method below.
    For most cases, use derived classes instead of this base class.

    Common operators (addition and multiplication by a scalar) are provided as part of this base class.
    For such operations, the user is responsible for handling the complex phase and weights of summed profiles.
    In particular, summing between different types of profiles is not recommended.
    This feature is tailored for linear combination of profiles of the same type, in particular analytical profiles like Laguerre-gauss or Hermite-Gauss.
    """

    def __init__(self):
        # Initialise x and y spatial offsets as placeholders
        self.x_offset = 0
        self.y_offset = 0
        self.is_plane_wave = False

    def _evaluate(self, x, y):
        """
        Return the transverse envelope.

        Parameters
        ----------
        x, y : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope: ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y
        """
        # The base class only defines dummy fields
        # (This should be replaced by any class that inherits from this one.)
        return np.zeros(x.shape, dtype="complex128")

    def __add__(self, other):
        """Return the sum of two transverse profiles."""
        return SummedTransverseProfile(self, other)

    def __mul__(self, factor):
        """Return the scaled transverse profile."""
        return ScaledTransverseProfile(self, factor)

    def __rmul__(self, factor):
        """Return the scaled transverse profile."""
        return ScaledTransverseProfile(self, factor)

    def __update_is_plane_wave__(self, value):
        """Update state of is_plane_wave variable."""
        self.is_plane_wave = value

    def evaluate(self, x, y):
        """
        Return the transverse envelope modified by any spatial offsets.

        Parameters
        ----------
        x, y: ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope: ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y
        """
        return self._evaluate(x + self.x_offset, y + self.y_offset)

    def set_offset(self, x_offset, y_offset):
        """
        Populate the x and y spatial offsets of the profile.

        The profile will be shifted by these according to
        x+x_offset and y+y_offset prior to execution of
        _evaluate.

        Parameters
        ----------
        x_offset, y_offset: floats (m)
            Define spatial offsets to the beam. That is, how much
            to shift the beam by transversely
        """
        self.x_offset = x_offset
        self.y_offset = y_offset

        return self


class SummedTransverseProfile(TransverseProfile):
    """
    Base class for transverse profiles that are the sum of several other transverse profiles.

    Transverse Profile class that represents the sum of multiple transverse profiles.

    Parameters
    ----------
    transverse_profiles: list of TransverseProfile objects
        List of transverse profiles to be summed.
    """

    def __init__(self, *transverse_profiles):
        """Initialize the summed profile."""
        TransverseProfile.__init__(self)
        # Check that all transverse_profiles are TransverseProfile objects
        assert all([isinstance(tp, TransverseProfile) for tp in transverse_profiles]), (
            "All summands must be Profile objects."
        )
        self.transverse_profiles = transverse_profiles

    def evaluate(self, x, y):
        """Return the envelope field of the summed profile."""
        # Sum the fields of each profile
        return sum([tp.evaluate(x, y) for tp in self.transverse_profiles])


class ScaledTransverseProfile(TransverseProfile):
    """
    Base class for transverse profiles that are scaled by a factor.

    Transverse Profile class that represents scaled transverse profiles.

    Parameters
    ----------
    transverse_profile: TrasnverseProfile object
        Trasnverse profile to be scaled.
    factor: int or float
        Factor by which to scale the profile.
    """

    def __init__(self, transverse_profile, factor):
        """Initialize the summed profile."""
        TransverseProfile.__init__(self)
        # Check that the factor is a number
        assert isinstance(factor, (int, float, complex)), "The factor must be a number."
        # Check that the profile is a Profile object
        assert isinstance(transverse_profile, TransverseProfile), (
            "The profile must be a TransverseProfile object."
        )
        self.transverse_profile = transverse_profile
        self.factor = factor

    def evaluate(self, x, y):
        """Return the envelope field of the scaled transverse profile."""
        # Sum the fields of each profile
        return self.transverse_profile.evaluate(x, y) * self.factor



import numpy as np

from .profile import Profile


class GaussianProfile(Profile):
    r"""
    Class for the analytic profile of a Gaussian laser pulse.

    More precisely, the electric field corresponds to:

    .. math::

        E_u(\boldsymbol{x}_\perp,t) = Re\left[ E_0\,
        \exp\left( -\frac{\boldsymbol{x}_\perp^2}{w_0^2}
        - \frac{(t-t_{peak})^2}{\tau^2} -i\omega_0(t-t_{peak})
        + i\phi_{cep}\right) \times p_u \right]

    where :math:`u` is either :math:`x` or :math:`y`, :math:`p_u` is
    the polarization vector, :math:`Re` represent the real part, and
    :math:`\boldsymbol{x}_\perp` is the transverse coordinate (orthogonal
    to the propagation direction). The other parameters in this formula
    are defined below.
    This profile also supports some chirp parameters that are omitted in the expression above for clarity.

    Parameters
    ----------
    wavelength : float (in meter)
        The main laser wavelength :math:`\lambda_0` of the laser, which
        defines :math:`\omega_0` in the above formula, according to
        :math:`\omega_0 = 2\pi c/\lambda_0`.

    pol : list of 2 complex numbers (dimensionless)
        Polarization vector. It corresponds to :math:`p_u` in the above
        formula ; :math:`p_x` is the first element of the list and
        :math:`p_y` is the second element of the list. Using complex
        numbers enables elliptical polarizations.

    laser_energy : float (in Joule)
        The total energy of the laser pulse. The amplitude of the laser
        field (:math:`E_0` in the above formula) is automatically
        calculated so that the pulse has the prescribed energy.

    w0 : float (in meter)
        The waist of the laser pulse, i.e. :math:`w_0` in the above formula.

    tau : float (in second)
        The duration of the laser pulse, i.e. :math:`\tau` in the above
        formula. Note that :math:`\tau = \tau_{FWHM}/\sqrt{2\log(2)}`,
        where :math:`\tau_{FWHM}` is the Full-Width-Half-Maximum duration
        of the intensity distribution of the pulse.

    t_peak : float (in second)
        The time at which the laser envelope reaches its maximum amplitude,
        i.e. :math:`t_{peak}` in the above formula.

    x0 , y0 : float (in meter) optional (default: '0')
        Transverse centroid for the laser pulse

    cep_phase : float (in radian), optional
        The Carrier Envelope Phase (CEP), i.e. :math:`\phi_{cep}`
        in the above formula (i.e. the phase of the laser
        oscillation, at the time where the laser envelope is maximum)

    z_foc : float (in meter), optional
        Position of the focal plane. (The laser pulse is initialized at `z=0`.)

    phi2 : float (in second^2), optional (default: '0')
        The group-delay dispersion defined as :math:`\phi^{(2)} = \frac{dt_0}{d\omega}`. Here :math:`t_0` is the temporal position of this frequency component.

    beta : float (in second), optional (default: '0')
        The angular dispersion defined as :math:`\beta = \frac{d\theta_0}{d\omega}`. Here :math:`\theta_0` is the propagation angle of this frequency component.

    zeta : float (in meter * second), optional (default: '0')
        A spatial chirp defined as :math:`\zeta = \frac{dx_0}{d\omega}`. Here :math:`x_0` is the transverse beam center position of this frequency component.
        The definitions of beta, phi2, and zeta are taken from [S. Akturk et al., Optics Express 12, 4399 (2004)].

    stc_theta : float (in radian), optional (default: '0')
        Transverse direction along which there are chirps and spatio-temporal couplings.
        A value of 0 corresponds to the x-axis.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from lasy.laser import Laser
    >>> from lasy.profiles.gaussian_profile import GaussianProfile
    >>> from lasy.utils.laser_utils import get_full_field
    >>> # Create profile.
    >>> profile = GaussianProfile(
    ...     wavelength=0.6e-6,  # m
    ...     pol=(1, 0),
    ...     laser_energy=1.0,  # J
    ...     w0=5e-6,  # m
    ...     tau=30e-15,  # s
    ...     t_peak=0.0,  # s
    ... )
    >>> # Create laser with given profile in `rt` geometry.
    >>> laser = Laser(
    ...     dim="rt",
    ...     lo=(0e-6, -60e-15),
    ...     hi=(10e-6, +60e-15),
    ...     npoints=(50, 400),
    ...     profile=profile,
    ... )
    >>> # Visualize field.
    >>> E_rt, extent = get_full_field(laser)
    >>> extent[2:] *= 1e6
    >>> extent[:2] *= 1e15
    >>> tmin, tmax, rmin, rmax = extent
    >>> vmax = np.abs(E_rt).max()
    >>> plt.imshow(
    ...     E_rt,
    ...     origin="lower",
    ...     aspect="auto",
    ...     vmax=vmax,
    ...     vmin=-vmax,
    ...     extent=[tmin, tmax, rmin, rmax],
    ...     cmap="bwr",
    ... )
    >>> plt.xlabel("t (fs)")
    >>> plt.ylabel("r (µm)")
    """

    def __init__(
        self,
        wavelength,
        pol,
        laser_energy,
        w0,
        tau,
        t_peak,
        x0=0,
        y0=0,
        cep_phase=0,
        z_foc=0,
        phi2=0,
        beta=0,
        zeta=0,
        stc_theta=0,
    ):
        super().__init__(wavelength, pol)
        self.laser_energy = laser_energy
        self.w0 = w0
        self.tau = tau
        self.t_peak = t_peak
        self.cep_phase = cep_phase
        self.z_foc = z_foc
        self.z_foc_over_zr = z_foc * wavelength / (np.pi * w0**2)
        self.phi2 = phi2
        self.beta = beta
        self.zeta = zeta
        self.stc_theta = stc_theta
        self.x0 = x0
        self.y0 = y0

    def evaluate(self, x, y, t):
        """
        Return the longitudinal envelope.

        Parameters
        ----------
        t : ndarrays of floats
            Define longitudinal points on which to evaluate the envelope

        x,y : ndarrays of floats
            Define transverse points on which to evaluate the envelope

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the longitudinal envelope at the
            specified points. This array has the same shape as the array t.
        """
        inv_tau2 = self.tau ** (-2)
        inv_complex_waist_2 = 1.0 / (
            self.w0**2 * (1.0 + 2.0j * self.z_foc / (self.k0 * self.w0**2))
        )
        stretch_factor = (
            1
            + 4.0
            * (-self.zeta + self.beta * self.z_foc)
            * inv_tau2
            * (-self.zeta + self.beta * self.z_foc)
            * inv_complex_waist_2
            + 2.0j * (-self.phi2 - self.beta**2 * self.k0 * self.z_foc) * inv_tau2
        )
        stc_exponent = (
            1.0
            / stretch_factor
            * inv_tau2
            * (
                t
                - self.t_peak
                - self.beta
                * self.k0
                * (x * np.cos(self.stc_theta) + y * np.sin(self.stc_theta))
                - 2.0j
                * (x * np.cos(self.stc_theta) + y * np.sin(self.stc_theta))
                * (-self.zeta - self.beta * self.z_foc)
                * inv_complex_waist_2
            )
            ** 2
        )
        # Term for wavefront curvature + Gouy phase
        diffract_factor = 1.0 - 1j * self.z_foc_over_zr
        # Calculate the argument of the complex exponential
        exp_argument = -((x - self.x0) ** 2 + (y - self.y0) ** 2) / (
            self.w0**2 * diffract_factor
        )
        # Get the profile
        envelope = (
            np.exp(
                -stc_exponent
                + exp_argument
                + 1.0j * (self.cep_phase + self.omega0 * self.t_peak)
            )
            / diffract_factor
        )

        return envelope



import numpy as np

from .longitudinal_profile import LongitudinalProfile


class CosineLongitudinalProfile(LongitudinalProfile):
    r"""
    Class for the analytic longitudinal truncated cosine profile of a laser pulse.

    More precisely, the longitudinal envelope
    (to be used in the :class:`.CombinedLongitudinalTransverseProfile` class)
    corresponds to:

    .. math::

        \mathcal{L}(t) = \cos\left({ \frac{\pi}{2} \frac{t-t_{peak}}{\tau_{fwhm}} }\right)
            \theta\left({ \frac{t-t_{peak}}{\tau_{fwhm}} + 1 }\right)
            \theta\left({ 1 - \frac{t-t_{peak}}{\tau_{fwhm}}} \right)
            \exp\left({ + i (\phi_{cep} + \omega_0 t_{peak} ) }\right)


    Parameters
    ----------
    wavelength : float (in meter)
        The main laser wavelength :math:`\lambda_0` of the laser.

    tau_fwhm : float (in second)
        The Full-Width-Half-Maximum duration of the intensity distribution of the pulse,
        i.e. :math:`\tau_{fwhm}` in the above formula.

    t_peak : float (in second)
        The time at which the laser envelope reaches its maximum amplitude,
        i.e. :math:`t_{peak}` in the above formula.

    cep_phase : float (in radian), optional
        The Carrier Enveloppe Phase (CEP)
        (i.e. the phase of the laser oscillation, at the time where the
        laser envelope is maximum,  :math:`\phi_{cep}` in the above formula).
    """

    def __init__(self, wavelength, tau_fwhm, t_peak, cep_phase=0):
        super().__init__(wavelength)
        self.tau_fwhm = tau_fwhm
        self.t_peak = t_peak
        self.cep_phase = cep_phase

    def evaluate(self, t):
        """
        Return the longitudinal envelope.

        Parameters
        ----------
        t : ndarrays of floats
            Define points on which to evaluate the envelope

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the longitudinal envelope at the
            specified points. This array has the same shape as the array t.
        """
        tn = (t - self.t_peak) / self.tau_fwhm

        envelope = (
            np.cos(0.5 * np.pi * tn)
            * (tn > -1)
            * (tn < 1)
            * np.exp(+1.0j * (self.cep_phase + self.omega0 * self.t_peak))
        )

        return envelope



# -*- coding: utf-8 -*-

import copy

import numpy as np
import pytest
from scipy.constants import c

from lasy.laser import Laser
from lasy.optical_elements.parabolic_mirror import ParabolicMirror
from lasy.profiles import (
    CombinedLongitudinalTransverseProfile,
    FromArrayProfile,
    GaussianProfile,
    SpeckleProfile,
)
from lasy.profiles.longitudinal import (
    CosineLongitudinalProfile,
    GaussianLongitudinalProfile,
    LongitudinalProfileFromData,
    SuperGaussianLongitudinalProfile,
)
from lasy.profiles.profile import Profile, ScaledProfile, SummedProfile
from lasy.profiles.transverse import (
    FlattenedGaussianTransverseProfile,
    GaussianTransverseProfile,
    HermiteGaussianTransverseProfile,
    JincTransverseProfile,
    LaguerreGaussianTransverseProfile,
    ScaledTransverseProfile,
    SummedTransverseProfile,
    SuperGaussianTransverseProfile,
    TransverseProfile,
    TransverseProfileFromData,
)
from lasy.utils.exp_data_utils import find_center_of_mass
from lasy.utils.grid import Grid
from lasy.utils.laser_utils import compute_laser_energy, get_w0


class MockProfile(Profile):
    """A mock Profile class that always returns a constant value."""

    def __init__(self, wavelength, pol, value):
        super().__init__(wavelength, pol)
        self.value = value

    def evaluate(self, x, y, t):
        return np.ones_like(x, dtype="complex128") * self.value


class MockTransverseProfile(TransverseProfile):
    """A mock TransverseProfile class that always returns a constant value."""

    def __init__(self, value):
        super().__init__()
        self.value = value

    def evaluate(self, x, y):
        return np.ones_like(x, dtype="complex128") * self.value


@pytest.fixture(scope="function")
def gaussian():
    # Cases with Gaussian laser
    wavelength = 0.8e-6
    pol = (1, 0)
    laser_energy = 1.0  # J
    t_peak = 0.0e-15  # s
    tau = 30.0e-15  # s
    w0 = 5.0e-6  # m
    profile = GaussianProfile(wavelength, pol, laser_energy, w0, tau, t_peak)

    return profile


def test_transverse_profiles_rt():
    npoints = 4000
    w0 = 10.0e-6
    r = np.linspace(0, 12 * w0, npoints)

    # GaussianTransverseProfile
    print("GaussianTransverseProfile")
    std_th = w0 / np.sqrt(2)
    profile = GaussianTransverseProfile(w0)
    field = profile.evaluate(r, np.zeros_like(r))
    std = np.sqrt(np.average(r**2, weights=np.abs(field)))
    print("std_th = ", std_th)
    print("std = ", std)
    assert np.abs(std - std_th) / std_th < 0.01

    # SuperGaussianTransverseProfile
    print("SuperGaussianTransverseProfile")
    n_order = 100  # close to flat-top, compared with flat-top theory
    std_th = w0 / np.sqrt(3)
    profile = SuperGaussianTransverseProfile(w0, n_order)
    field = profile.evaluate(r, np.zeros_like(r))
    std = np.sqrt(np.average(r**2, weights=np.abs(field)))
    print("std_th = ", std_th)
    print("std = ", std)
    assert np.abs(std - std_th) / std_th < 0.01

    # JincTransverseProfile
    print("JincTransverseProfile")
    profile = JincTransverseProfile(w0)
    std_th = 1.5 * w0  # Just measured from this test
    field = profile.evaluate(r, np.zeros_like(r))
    std = np.sqrt(np.average(r**2, weights=abs(field) ** 2))
    print("std_th = ", std_th)
    print("std = ", std)
    assert np.abs(std - std_th) / std_th < 0.1

    # LaguerreGaussianTransverseProfile
    print("LaguerreGaussianTransverseProfile")
    p = 2
    m = 0
    std_th = np.sqrt((2 * p + m + 1) / 2) * w0
    profile = LaguerreGaussianTransverseProfile(w0, p, m, wavelength=800e-9)
    field = profile.evaluate(r, np.zeros_like(r))
    std = np.sqrt(np.average(r**2, weights=np.abs(r) * np.abs(field) ** 2))
    print("std_th = ", std_th)
    print("std = ", std)
    assert np.abs(std - std_th) / std_th < 0.01


def test_transverse_profiles_3d():
    w0_x = 10.0e-6
    w0_y = 12.0e-6

    # HermiteGaussianTransverseProfile
    print("HermiteGaussianTransverseProfile")
    m = 2
    n = 2
    std_th_x = np.sqrt(5.0 / 4) * w0_x
    std_th_y = np.sqrt(5.0 / 4) * w0_y
    profile = HermiteGaussianTransverseProfile(w0_x, w0_y, m, n, wavelength=800e-9)
    x = np.linspace(-4 * w0_x, 4 * w0_x, 200)
    y = np.linspace(-4 * w0_y, 4 * w0_y, 150)
    field_x = profile.evaluate(x, np.zeros_like(x))
    field_y = profile.evaluate(np.zeros_like(y), y)
    std_x = np.sqrt(np.average(x**2, weights=np.abs(field_x) ** 2))
    std_y = np.sqrt(np.average(y**2, weights=np.abs(field_y) ** 2))
    print("std_th_x = ", std_th_x)
    print("std_x = ", std_x)
    print("std_th_y = ", std_th_y)
    print("std_y = ", std_y)
    assert np.abs(std_x - std_th_x) / std_th_x < 0.01
    assert np.abs(std_y - std_th_y) / std_th_y < 0.01

    # TransverseProfileFromData
    print("TransverseProfileFromData")
    lo = (-40.0e-6, -40.0e-6)
    hi = (40.0e-6, 40.0e-6)
    x = np.linspace(lo[0], hi[0], 200)
    y = np.linspace(lo[1], hi[1], 100)
    dx = x[1] - x[0]
    w0 = 10.0e-6
    x0 = 10.0e-6
    X, Y = np.meshgrid(x, y, indexing="ij")
    intensity_data = np.exp(-((X - x0) ** 2 + Y**2) / w0**2)
    profile = TransverseProfileFromData(intensity_data, lo, hi)
    field = profile.evaluate(X, Y)
    x0_test = find_center_of_mass(field**2)[0] * dx + lo[0]
    print("beam center, theory: ", x0)
    print("beam center from profile ", x0_test)
    assert (x0_test - x0) / x0 < 0.1


def test_longitudinal_profiles():
    npoints = 10000

    wavelength = 800e-9
    tau_fwhm = 30.0e-15
    omega_fwhm = 4 * np.log(2) / tau_fwhm  # Assumes fully-compressed
    t_peak = 1.0 * tau_fwhm
    cep_phase = 0.5 * np.pi
    omega_0 = 2.0 * np.pi * c / wavelength

    t = np.linspace(t_peak - 4 * tau_fwhm, t_peak + 4 * tau_fwhm, npoints)
    omega = np.linspace(omega_0 - 4 * omega_fwhm, omega_0 + 4 * omega_fwhm, npoints)
    wavelength_axis = 2.0 * np.pi * c / omega  # Note: monotonically decreasing

    # GaussianLongitudinalProfile
    print("GaussianLongitudinalProfile")
    tau = tau_fwhm / np.sqrt(2 * np.log(2))
    profile_gaussian = GaussianLongitudinalProfile(wavelength, tau, t_peak, cep_phase)
    field_gaussian = profile_gaussian.evaluate(t)

    std_gauss = np.sqrt(np.average((t - t_peak) ** 2, weights=np.abs(field_gaussian)))
    std_gauss_th = tau / np.sqrt(2.0)
    print("std_th = ", std_gauss_th)
    print("std = ", std_gauss)
    assert np.abs(std_gauss - std_gauss_th) / std_gauss_th < 0.01

    t_peak_gaussian = t[np.argmax(np.abs(field_gaussian))]
    print("t_peak_th = ", t_peak)
    print("t_peak = ", t_peak_gaussian)
    assert np.abs(t_peak_gaussian - t_peak) / t_peak < 0.01

    ff_gaussian = field_gaussian * np.exp(-1.0j * omega_0 * t)
    cep_phase_gaussian = np.angle(ff_gaussian[np.argmax(np.abs(field_gaussian))])
    print("cep_phase_th = ", cep_phase)
    print("cep_phase = ", cep_phase_gaussian)
    assert np.abs(cep_phase_gaussian - cep_phase) / cep_phase < 0.02

    # SuperGaussianLongitudinalProfile
    print("SuperGaussianLongitudinalProfile")
    n_order = 2  # ordinary gaussian
    tau = tau_fwhm / np.sqrt(2 * np.power(np.log(2), n_order / 2))
    profile_super_gaussian = SuperGaussianLongitudinalProfile(
        wavelength, tau, t_peak, n_order, cep_phase
    )
    field_super_gaussian = profile_super_gaussian.evaluate(t)

    std_super_gauss = np.sqrt(
        np.average((t - t_peak) ** 2, weights=np.abs(field_super_gaussian))
    )
    std_super_gauss_th = tau / np.sqrt(2.0)
    print("std_th = ", std_super_gauss_th)
    print("std = ", std_super_gauss)
    assert np.abs(std_super_gauss - std_super_gauss_th) / std_super_gauss_th < 0.01

    t_peak_super_gaussian = t[np.argmax(np.abs(field_super_gaussian))]
    print("t_peak_th = ", t_peak)
    print("t_peak = ", t_peak_super_gaussian)
    assert np.abs(t_peak_super_gaussian - t_peak) / t_peak < 0.01

    ff_super_gaussian = field_super_gaussian * np.exp(-1.0j * omega_0 * t)
    cep_phase_super_gaussian = np.angle(
        ff_super_gaussian[np.argmax(np.abs(field_super_gaussian))]
    )
    print("cep_phase_th = ", cep_phase)
    print("cep_phase = ", cep_phase_super_gaussian)
    assert np.abs(cep_phase_super_gaussian - cep_phase) / cep_phase < 0.02

    # CosineLongitudinalProfile
    print("CosineLongitudinalProfile")
    profile_cos = CosineLongitudinalProfile(wavelength, tau_fwhm, t_peak, cep_phase)
    field_cos = profile_cos.evaluate(t)

    std_cos = np.sqrt(np.average((t - t_peak) ** 2, weights=np.abs(field_cos)))
    std_cos_th = tau_fwhm * np.sqrt(1 - 8 / np.pi**2)
    print("std_th = ", std_cos_th)
    print("std = ", std_cos)
    assert np.abs(std_cos - std_cos_th) / std_cos_th < 0.01

    t_peak_cos = t[np.argmax(np.abs(field_cos))]
    print("t_peak_th = ", t_peak)
    print("t_peak = ", t_peak_cos)
    assert np.abs(t_peak_cos - t_peak) / t_peak < 0.01

    ff_cos = field_cos * np.exp(-1.0j * omega_0 * t)
    cep_phase_cos = np.angle(ff_cos[np.argmax(np.abs(field_cos))])
    print("cep_phase_th = ", cep_phase)
    print("cep_phase = ", cep_phase_cos)
    assert np.abs(cep_phase_cos - cep_phase) / cep_phase < 0.02

    # LongitudinalProfileFromData
    print("LongitudinalProfileFromData")
    data = {}  # Generate spectral data assuming analytic Fourier transform of GaussianLongitudinalProfile
    data["datatype"] = "spectral"
    data["dt"] = 1e-16
    profile = np.exp(
        -(tau**2) * ((omega - omega_0) ** 2) / 4.0 + 1.0j * (cep_phase + omega * t_peak)
    )
    spectral_intensity = np.abs(profile) ** 2 / np.max(np.abs(profile) ** 2)
    spectral_phase = np.unwrap(np.angle(profile))

    print("Case 1: monotonically decreasing data on wavelength axis")
    data["axis"] = wavelength_axis
    data["intensity"] = spectral_intensity
    data["phase"] = spectral_phase
    profile_data = LongitudinalProfileFromData(data, np.min(t), np.max(t))
    field_data = profile_data.evaluate(t)

    std_gauss_data = np.sqrt(np.average((t - t_peak) ** 2, weights=np.abs(field_data)))
    std_gauss_th = tau / np.sqrt(2.0)
    print("std_th = ", std_gauss_th)
    print("std = ", std_gauss_data)
    assert np.abs(std_gauss_data - std_gauss_th) / std_gauss_th < 0.01

    t_peak_gaussian_data = t[np.argmax(np.abs(field_data))]
    print("t_peak_th = ", t_peak)
    print("t_peak = ", t_peak_gaussian_data)
    assert np.abs(t_peak_gaussian_data - t_peak) / t_peak < 0.01

    print("Case 2: monotonically increasing data on wavelength axis")
    data["axis"] = wavelength_axis[::-1]
    data["intensity"] = spectral_intensity[::-1]
    data["phase"] = spectral_phase[::-1]
    profile_data = LongitudinalProfileFromData(data, np.min(t), np.max(t))
    field_data = profile_data.evaluate(t)

    std_gauss_data = np.sqrt(np.average((t - t_peak) ** 2, weights=np.abs(field_data)))
    std_gauss_th = tau / np.sqrt(2.0)
    print("std_th = ", std_gauss_th)
    print("std = ", std_gauss_data)
    assert np.abs(std_gauss_data - std_gauss_th) / std_gauss_th < 0.01

    t_peak_gaussian_data = t[np.argmax(np.abs(field_data))]
    print("t_peak_th = ", t_peak)
    print("t_peak = ", t_peak_gaussian_data)
    assert np.abs(t_peak_gaussian_data - t_peak) / t_peak < 0.01

    print("Case 3: monotonically increasing data on angular frequency axis")
    data["axis"] = omega
    data["intensity"] = spectral_intensity
    data["phase"] = spectral_phase
    data["axis_is_wavelength"] = False
    profile_data = LongitudinalProfileFromData(data, np.min(t), np.max(t))
    field_data = profile_data.evaluate(t)

    std_gauss_data = np.sqrt(np.average((t - t_peak) ** 2, weights=np.abs(field_data)))
    std_gauss_th = tau / np.sqrt(2.0)
    print("std_th = ", std_gauss_th)
    print("std = ", std_gauss_data)
    assert np.abs(std_gauss_data - std_gauss_th) / std_gauss_th < 0.01

    t_peak_gaussian_data = t[np.argmax(np.abs(field_data))]
    print("t_peak_th = ", t_peak)
    print("t_peak = ", t_peak_gaussian_data)
    assert np.abs(t_peak_gaussian_data - t_peak) / t_peak < 0.01

    print("Case 4: monotonically decreasing data on angular frequency axis")
    data["axis"] = omega[::-1]
    data["intensity"] = spectral_intensity[::-1]
    data["phase"] = spectral_phase[::-1]
    data["axis_is_wavelength"] = False
    profile_data = LongitudinalProfileFromData(data, np.min(t), np.max(t))
    field_data = profile_data.evaluate(t)

    std_gauss_data = np.sqrt(np.average((t - t_peak) ** 2, weights=np.abs(field_data)))
    std_gauss_th = tau / np.sqrt(2.0)
    print("std_th = ", std_gauss_th)
    print("std = ", std_gauss_data)
    assert np.abs(std_gauss_data - std_gauss_th) / std_gauss_th < 0.01

    t_peak_gaussian_data = t[np.argmax(np.abs(field_data))]
    print("t_peak_th = ", t_peak)
    print("t_peak = ", t_peak_gaussian_data)
    assert np.abs(t_peak_gaussian_data - t_peak) / t_peak < 0.01


def test_profile_gaussian_3d_cartesian(gaussian):
    # - 3D Cartesian case
    dim = "xyt"
    lo = (-10e-6, -10e-6, -60e-15)
    hi = (+10e-6, +10e-6, +60e-15)
    npoints = (100, 100, 100)

    laser = Laser(dim, lo, hi, npoints, gaussian)
    laser.write_to_file("gaussianlaser3d")
    laser.propagate(1e-6)
    laser.write_to_file("gaussianlaser3d")


def test_profile_gaussian_cylindrical(gaussian):
    # - Cylindrical case
    dim = "rt"
    lo = (0e-6, -60e-15)
    hi = (10e-6, +60e-15)
    npoints = (50, 100)

    laser = Laser(dim, lo, hi, npoints, gaussian)
    laser.write_to_file("gaussianlaserRZ")
    laser.propagate(1e-6)
    laser.write_to_file("gaussianlaserRZ")


def test_from_array_profile():
    # Create a 3D numpy array, use it to create a LASY profile,
    # and check that the resulting profile has the correct width
    lo = (-10e-6, -20e-6, -30e-15)
    hi = (+10e-6, +20e-6, +30e-15)
    npoints = (16, 32, 64)
    x = np.linspace(lo[0], hi[0], npoints[0])
    y = np.linspace(lo[1], hi[1], npoints[1])
    t = np.linspace(lo[2], hi[2], npoints[2])
    X, Y, T = np.meshgrid(x, y, t, indexing="ij")
    e0 = 4.0e12  # Approx a0 = 1 for wavelength = 800 nm
    wx = 3.0e-6
    wy = 5.0e-6
    tau = 5.0e-15
    E = 1j * e0 * np.exp(-(X**2) / wx**2 - Y**2 / wy**2 - t**2 / tau**2)
    wavelength = 0.8e-6
    pol = (1, 0)
    axes = {"x": x, "y": y, "t": t}
    dim = "xyt"

    profile = FromArrayProfile(
        wavelength=wavelength, pol=pol, array=E, dim=dim, axes=axes
    )
    laser = Laser(dim, lo, hi, npoints, profile)
    laser.write_to_file("fromArray")

    F = profile.evaluate(X, Y, T)
    width = (
        np.sqrt(
            np.sum(np.abs(F) ** 2 * x.reshape((x.size, 1, 1)) ** 2)
            / np.sum(np.abs(F) ** 2)
        )
        * 2
    )
    print("theory width  : ", wx)
    print("Measured width: ", width)
    assert np.abs((width - wx) / wx) < 1.0e-5


def test_speckle_profile():
    # - speckled laser case
    print("SpeckledProfile")
    wavelength = 0.351e-6  # Laser wavelength in meters
    polarization = (1, 0)  # Linearly polarized in the x direction
    laser_energy = 1.0  # J (this is the laser energy stored in the box defined by `lo` and `hi` below)
    focal_length = 3.5  # m
    beam_aperture = [0.35, 0.5]  # m
    n_beamlets = [24, 32]
    temporal_smoothing_type = "GP ISI"
    relative_laser_bandwidth = 0.005

    profile = SpeckleProfile(
        wavelength,
        polarization,
        laser_energy,
        focal_length,
        beam_aperture,
        n_beamlets,
        temporal_smoothing_type=temporal_smoothing_type,
        relative_laser_bandwidth=relative_laser_bandwidth,
    )
    dimensions = "xyt"
    dx = wavelength * focal_length / beam_aperture[0]
    dy = wavelength * focal_length / beam_aperture[1]
    Lx = 1.8 * dx * n_beamlets[0]
    Ly = 3.1 * dy * n_beamlets[1]
    nu_laser = c / wavelength
    t_max = 50 / nu_laser
    lo = (0, 0, 0)
    hi = (Lx, Ly, t_max)
    npoints = (200, 250, 2)

    laser = Laser(dimensions, lo, hi, npoints, profile)
    laser.write_to_file("speckledProfile")


def test_add_profiles():
    # Add the two profiles together
    profile_1 = MockProfile(0.8e-6, (1, 0), 1.0)
    profile_2 = MockProfile(0.8e-6, (1, 0), 2.0)
    summed_profile = profile_1 + profile_2
    # Check that the result is a SummedProfile object
    assert isinstance(summed_profile, SummedProfile)
    # Check that the profiles are stored correctly
    assert summed_profile.profiles[0] == profile_1
    assert summed_profile.profiles[1] == profile_2
    # Check that the evaluate method works
    assert np.allclose(summed_profile.evaluate(0, 0, 0), 3.0)


def test_add_error_if_not_all_profiles():
    profile_1 = MockProfile(0.8e-6, (1, 0), 1.0)
    with pytest.raises(AssertionError):
        profile_1 + 1.0


def test_add_error_if_different_wavelength():
    profile_1 = MockProfile(0.8e-6, (1, 0), 1.0)
    profile_2 = MockProfile(0.8e-6, (1, 0), 2.0)
    profile_3 = MockProfile(0.9e-6, (1, 0), 2.0)
    summed_profile = profile_1 + profile_2
    with pytest.raises(AssertionError):
        summed_profile + profile_3


def test_add_error_if_different_polarisation():
    profile_1 = MockProfile(0.8e-6, (1, 0), 1.0)
    profile_2 = MockProfile(0.8e-6, (1, 0), 2.0)
    profile_3 = MockProfile(0.8e-6, (0, 1), 2.0)
    summed_profile = profile_1 + profile_2
    with pytest.raises(AssertionError):
        summed_profile + profile_3


def test_scale_profiles():
    # Add the two profiles together
    profile_1 = MockProfile(0.8e-6, (1, 0), 1.0)
    scaled_profile = 2.0 * profile_1
    scaled_profile_right = profile_1 * 2.0
    # Check that the result is a ScaledProfile object
    assert isinstance(scaled_profile, ScaledProfile)
    # Check that the profiles are stored correctly
    assert scaled_profile.profile == profile_1
    # Check that the evaluate method works
    assert np.allclose(scaled_profile.evaluate(0, 0, 0), 2.0)
    assert np.allclose(scaled_profile_right.evaluate(0, 0, 0), 2.0)


def test_scale_error_if_not_scalar():
    profile_1 = MockProfile(0.8e-6, (1, 0), 1.0)
    with pytest.raises(AssertionError):
        profile_1 * profile_1
    with pytest.raises(AssertionError):
        profile_1 * [1.0, 2.0]


def test_add_transverse_profiles():
    # Add the two profiles together
    trans_profile_1 = MockTransverseProfile(1.0)
    trans_profile_2 = MockTransverseProfile(2.0)
    summed_trans_profile = trans_profile_1 + trans_profile_2
    # Check that the result is a SummedTransverseProfile object
    assert isinstance(summed_trans_profile, SummedTransverseProfile)
    # Check that the profiles are stored correctly
    assert summed_trans_profile.transverse_profiles[0] == trans_profile_1
    assert summed_trans_profile.transverse_profiles[1] == trans_profile_2
    # Check that the evaluate method works
    assert np.allclose(summed_trans_profile.evaluate(0, 0), 3.0)


def test_add_transverse_error_if_not_all_transverse_profiles():
    trans_profile_1 = MockTransverseProfile(1.0)
    with pytest.raises(AssertionError):
        trans_profile_1 + 1.0


def test_scale_transverse_profiles():
    # Add the two profiles together
    trans_profile_1 = MockTransverseProfile(1.0)
    scaled_trans_profile = 2.0 * trans_profile_1
    scaled_trans_profile_right = trans_profile_1 * 2.0
    # Check that the result is a ScaledProfile object
    assert isinstance(scaled_trans_profile, ScaledTransverseProfile)
    assert isinstance(scaled_trans_profile_right, ScaledTransverseProfile)
    # Check that the profiles are stored correctly
    assert scaled_trans_profile.transverse_profile == trans_profile_1
    # Check that the evaluate method works
    assert np.allclose(scaled_trans_profile.evaluate(0, 0), 2.0)
    assert np.allclose(scaled_trans_profile.evaluate(0, 0), 2.0)


def test_scale_trans_error_if_not_scalar():
    trans_profile_1 = MockTransverseProfile(1.0)
    with pytest.raises(AssertionError):
        trans_profile_1 * trans_profile_1
    with pytest.raises(AssertionError):
        trans_profile_1 * [1.0, 2.0]


def test_flattened_gaussian_profile():
    w = 20e-3
    N = 25
    wl = 800e-9
    tau = 30e-15
    pol = (1, 0)
    energy = 1.0
    focal_length = 1.0

    w0 = focal_length * wl / np.pi / w

    nf = FlattenedGaussianTransverseProfile(
        field_type="nearfield", w=w, N=N, wavelength=wl
    )
    ff = FlattenedGaussianTransverseProfile(
        field_type="farfield", w=w0, N=N, wavelength=wl
    )

    long = GaussianLongitudinalProfile(wl, tau, 0)

    nf_prof = CombinedLongitudinalTransverseProfile(
        wl, pol, long, nf, laser_energy=energy
    )
    ff_prof = CombinedLongitudinalTransverseProfile(
        wl, pol, long, ff, laser_energy=energy
    )

    dim = "rt"
    lo = (0, -100e-15)
    hi_ff = (500e-6, 100e-15)
    hi_nf = (40e-3, 100e-15)
    npoints = (5000, 200)

    las_nf = Laser(dim, lo, hi_nf, npoints, nf_prof)
    las_ff = Laser(dim, lo, hi_ff, npoints, ff_prof)

    las_nf_cp = copy.deepcopy(las_nf)

    OAP = ParabolicMirror(f=focal_length)
    las_nf_cp.apply_optics(OAP)
    las_nf_cp.propagate(
        focal_length, grid=Grid(dim, lo, hi_ff, npoints, n_azimuthal_modes=1)
    )

    radlineout_nf_cp = (
        np.abs(las_nf_cp.grid.get_temporal_field()[0, :, int(npoints[1] / 2)]) ** 2
    )
    radlineout_ff = (
        np.abs(las_ff.grid.get_temporal_field()[0, :, int(npoints[1] / 2)]) ** 2
    )

    err = np.sum(
        np.abs(np.abs(radlineout_nf_cp) ** 2 - np.abs(radlineout_ff) ** 2)
    ) / np.sum(np.abs(radlineout_ff) ** 2)

    assert err < 1e-2

    energy_ff = compute_laser_energy(dim, las_ff.grid)
    energy_nf_cp = compute_laser_energy(dim, las_nf_cp.grid)
    assert np.abs(energy_ff - energy_nf_cp) / energy_ff < 1e-4

    w0_ff = get_w0(las_ff.grid, dim)
    w0_nf_cp = get_w0(las_nf_cp.grid, dim)
    assert np.abs(w0_nf_cp - w0_ff) / w0_ff < 1e-2



import numpy as np
from scipy.interpolate import RegularGridInterpolator

from .profile import Profile


class FromArrayProfile(Profile):
    r"""
    Profile defined from numpy array directly.

    The numpy array contains the envelope of the electric field of the laser pulse, defined as :math:`\mathcal{E}` in:

    .. math::
        \begin{aligned}
        E_x(x,y,t) = \operatorname{Re} \left( \mathcal{E}(x,y,t) e^{-i \omega_0t}p_x \right)\\
        E_y(x,y,t) = \operatorname{Re} \left( \mathcal{E}(x,y,t) e^{-i \omega_0t}p_y \right)\end{aligned}

    where :math:`\operatorname{Re}` stands for real part, :math:`E_x` (resp. :math:`E_y`) is the laser electric field in the :math:`x` (resp. :math:`y`) direction.

    Parameters
    ----------
    wavelength : float (in meter)
        The main laser wavelength :math:`\lambda_0` of the laser, which
        defines :math:`\omega_0` in the above formula, according to
        :math:`\omega_0 = 2\pi c/\lambda_0`.

    pol : list of 2 complex numbers (dimensionless)
        Polarization vector. It corresponds to :math:`p_u` in the above
        formula ; :math:`p_x` is the first element of the list and
        :math:`p_y` is the second element of the list. Using complex
        numbers enables elliptical polarizations.

    array : 3darray of complex numbers
        Contains the values of the envelope, defined as :math:`\mathcal{E}` in the above formula.

    dim : string
        "xyt" or "rt"

    axes : Python dictionary containing the axes vectors.
        Keys are 'x', 'y', 't'.
        Values are the 1D arrays of each axis.
        array.shape = (axes['x'].size, axes['y'].size, axes['t'].size) in 3D,
        and similar in cylindrical geometry.

    axes_order : List of strings
        Gives the name and ordering of the axes in the array.
        Currently, only implemented for 3D, and supported values are
        ['x', 'y', 't'] and ['t', 'y', 'x'].
    """

    def __init__(self, wavelength, pol, array, dim, axes, axes_order=["x", "y", "t"]):
        super().__init__(wavelength, pol)

        assert dim in ["xyt", "rt"]
        assert array.ndim == 3, "array must be 3D, [x,y,t] or [modes,r,t]."
        self.axes = axes
        self.dim = dim
        self.array = array

        if dim == "xyt":
            assert axes_order == ["x", "y", "t"]

            self.combined_field_interp = RegularGridInterpolator(
                (axes["x"], axes["y"], axes["t"]),
                np.abs(self.array) + 1.0j * np.unwrap(np.angle(self.array), axis=-1),
                bounds_error=False,
                fill_value=0.0,
            )
        else:  # dim = "rt"
            assert axes_order == ["r", "t"]

            # If the first point of radial axis is not 0, we "mirror" it,
            # to make correct interpolation within the first cell
            if axes["r"][0] != 0.0:
                # add mirrored point to the axis
                r = np.concatenate(([-axes["r"][0]], axes["r"]))
                # takes first element of the array in the radial dimension
                subarray = self.array[:, 0, :]
                # add it at the beginning to be the value at the mirrored point
                self.array = np.concatenate(
                    (subarray[:, np.newaxis, :], self.array), axis=1
                )
            else:
                r = axes["r"]

            self.field_interp_modes = []
            # Loop over the 2*m-1 elements of the array and create a separate
            # interpolator object for each of them.
            # Note that the field_interp_modes is not directly the complex
            # envelope because interpolating separately real and imag is not
            # accurate enough. Instead, the real part of field_interp_modes
            # represents the mode's modulus and its imag the mode's phase.
            for imode in range(self.array.shape[0]):
                self.field_interp_modes.append(
                    RegularGridInterpolator(
                        (r, axes["t"]),
                        np.abs(self.array[imode, :, :])
                        + 1.0j * np.unwrap(np.angle(self.array[imode, :, :]), axis=-1),
                        bounds_error=False,
                        fill_value=0.0,
                    )
                )

    def evaluate(self, x, y, t):
        """Return the envelope field of the scaled profile."""
        if self.dim == "xyt":
            combined_field = self.combined_field_interp((x, y, t))
        else:
            r = np.sqrt(x**2 + y**2)
            theta = np.angle(x + 1j * y)
            combined_field = np.zeros_like(x, dtype="complex128")
            nmodes = (len(self.field_interp_modes) + 1) // 2
            for imode in range(-nmodes + 1, nmodes):
                combined_field += self.field_interp_modes[imode]((r, t)) * np.exp(
                    -1j * imode * theta
                )

        return np.abs(np.real(combined_field)) * np.exp(1.0j * np.imag(combined_field))

    def evaluate_mrt(self, mode, r, t):
        """Return the envelope field of the scaled profile."""
        assert self.dim == "rt"
        combined_field = self.field_interp_modes[mode]((r, t))
        return np.abs(np.real(combined_field)) * np.exp(1.0j * np.imag(combined_field))



import numpy as np

from .transverse_profile import TransverseProfile


class GaussianTransverseProfile(TransverseProfile):
    r"""
    Class for the analytic profile of a Gaussian laser pulse.

    More precisely, at focus (``z_foc=0``), the transverse envelope
    (to be used in the :class:`.CombinedLongitudinalTransverseLaser` class)
    corresponds to:

    .. math::

        \mathcal{T}(x, y) = \exp\left( -\frac{x^2 + y^2}{w_0^2} \right)

    Parameters
    ----------
    w0 : float (in meter)
        The waist of the laser pulse, i.e. :math:`w_0` in the above formula.

    wavelength : float (in meter), optional
        The main laser wavelength :math:`\lambda_0` of the laser.
        (Only needed if ``z_foc`` is different than 0.)

    z_foc : float (in meter), optional
        Position of the focal plane. (The laser pulse is initialized at
        ``z=0``.)

    Warnings
    --------
    In order to initialize the pulse out of focus, you can either:

    - Use a non-zero ``z_foc``
    - Use ``z_foc=0`` (i.e. initialize the pulse at focus) and then call
      ``laser.propagate(-z_foc)``

    Both methods are in principle equivalent, but note that the first
    method uses the paraxial approximation, while the second method does
    not make this approximation.
    """

    def __init__(self, w0, wavelength=None, z_foc=0):
        super().__init__()
        self.w0 = w0
        if z_foc == 0:
            self.z_foc_over_zr = 0
        else:
            assert wavelength is not None, (
                "You need to pass the wavelength, when `z_foc` is non-zero."
            )
            self.z_foc_over_zr = z_foc * wavelength / (np.pi * w0**2)

    def _evaluate(self, x, y):
        """
        Return the transverse envelope.

        Parameters
        ----------
        x, y : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y
        """
        # Term for wavefront curvature + Gouy phase
        diffract_factor = 1.0 - 1j * self.z_foc_over_zr
        # Calculate the argument of the complex exponential
        exp_argument = -(x**2 + y**2) / (self.w0**2 * diffract_factor)
        # Get the transverse profile
        envelope = np.exp(exp_argument) / diffract_factor

        return envelope



import numpy as np

from lasy.utils.zernike import zernike

from .optical_element import OpticalElement


class ZernikeAberrations(OpticalElement):
    r"""
    Class for an optic applying a set of Zernike aberrations.

    More precisely, the amplitude multiplier corresponds to:

    .. math::

        T(\boldsymbol{x}_\perp,\omega) = \exp( i \sum_j a_j Z_j(\boldsymbol{x}_\perp))

    where
    :math:`\boldsymbol{x}_\perp` is the transverse coordinate (orthogonal
    to the propagation direction). :math:`Z_j` is the j-th Zernike polynomial,
    ordered according the OSA/ANSI indexing. The Zernike polynomials are normalized
    such that their integral over the unit disk is equal to :math:`\pi`. In the above
    formula, the total phase added to the pulse is a weighted sum of these Zernike
    polynomials with weights :math:`a_j`.

    For more information see: https://en.wikipedia.org/wiki/Zernike_polynomials


    Parameters
    ----------
    pupil_coords : tuple of floats (meters)
        A tuple of floats (cgx,cgy,r) with the first two elements corresponding to the center
        point and third element the radius of the pupil on which the zernike polynomial is
        defined.
    zernike_amplitudes : dict
        A dictionary with integer keys representing the OSA/ANSI indexing of the
        individual Zernike Polynomials. The values corresponding to these keys
        are floats giving the amplitudes / weights of the relevant Zernike polynomials.

    """

    def __init__(self, pupil_coords, zernike_amplitudes):
        self.pupil_coords = pupil_coords
        self.zernike_amplitudes = zernike_amplitudes

    def amplitude_multiplier(self, x, y, omega):
        """
        Return the amplitude multiplier.

        Parameters
        ----------
        x, y, omega : ndarrays of floats
            Define points on which to evaluate the multiplier.
            These arrays need to all have the same shape.

        Returns
        -------
        multiplier : ndarray of complex numbers
            Contains the value of the multiplier at the specified points.
            This array has the same shape as the array omega.
        """
        rr = np.sqrt(x**2 + y**2)
        phase = np.zeros_like(rr)

        for j in list(self.zernike_amplitudes):
            # Create the zernike phase and ensure it has the same number of dimensions as phase
            zernike_phase = zernike(x[..., 0], y[..., 0], self.pupil_coords, j)[
                ..., None
            ]  # Expand last axis

            # Increase the length of the frequency dimension such that the shape is suitable to be added
            # to the phase array, then add it
            phase += self.zernike_amplitudes[j] * np.broadcast_to(
                zernike_phase, phase.shape
            )

        return np.exp(1j * phase)



import numpy as np
from scipy.constants import c

from .profile import Profile


def gen_gaussian_time_series(t_num, dt, fwhm, rms_mean):
    """Generate a discrete time series that has gaussian power spectrum.

    Parameters
    ----------
    t_num : int
        The number of grid points in time

    fwhm : float
        The full width half maximum of the power spectrum

    rms_mean : scalar
        The root-mean-square average of the spectrum

    Returns
    -------
    temporal_amplitude : 1darray
        A time series array of complex numbers with shape [t_num]
    """
    if fwhm == 0.0:
        temporal_amplitude = np.zeros(t_num, dtype=np.complex128)
    else:
        omega = np.fft.fftshift(np.fft.fftfreq(t_num, d=dt))
        psd = np.exp(-np.log(2) * 0.5 * np.square(omega / fwhm * 2 * np.pi))
        spectral_amplitude = np.array(psd) * (
            np.random.normal(size=t_num) + 1j * np.random.normal(size=t_num)
        )
        temporal_amplitude = np.fft.ifftshift(
            np.fft.fft(np.fft.fftshift(spectral_amplitude))
        )
        temporal_amplitude *= rms_mean / np.sqrt(
            np.mean(np.square(np.abs(temporal_amplitude)))
        )
    return temporal_amplitude


class SpeckleProfile(Profile):
    r"""
    Class for the profile of a speckled laser pulse.

    Speckled lasers are used to mitigate laser-plasma interactions in fusion and ion acceleration contexts.
    More on the subject can be found in chapter 9 of `P. Michel, Introduction to Laser-Plasma Interactions <https://link.springer.com/book/10.1007/978-3-031-23424-8>`__.
    A speckled laser beam is a laser that is deliberately divided transversely into :math:`N_{bx}\times N_{by}` beamlets in the near-field.
    The phase plate provides a different phase to each beamlet, with index :math:`ml`, which then propagate to the far field and combine incoherently.

    The electric field in the focal plane, as a function of time :math:`t` and the coordinates
    :math:`\boldsymbol{x}_\perp=(x,y)` transverse to the direction of propagation, is:

    .. math::

        \begin{aligned}
        E_u(\boldsymbol{x}_\perp,t) &= Re\left[ E_0
        {\rm sinc}\left(\frac{\pi x}{\Delta x}\right)
        {\rm sinc}\left(\frac{\pi y}{\Delta y}\right)\times p_u
        \right.
        \\
        & \times\sum_{m,l=1}^{N_{bx}, N_{by}} A_{ml}
        \exp\left(i\boldsymbol{k}_{\perp ml}\cdot\boldsymbol{x}_\perp
        + i\phi_{{\rm RPP/CPP},ml}+i\psi_{{\rm SSD},ml}(t)\right)
        \Bigg]
        \end{aligned}

    where :math:`u` is either :math:`x` or :math:`y`, :math:`p_u` is
    the polarization vector, and :math:`Re` represent the real part [Michel, Eqns. 9.11, 87, 94].
    Several quantities are computed internally to the code depending on the
    method of smoothing chosen, including the beamlet amplitude :math:`A_{ml}`,
    the beamlet wavenumber :math:`k_{\perp ml}`,
    the relative phase contribution :math:`\phi_{{\rm RPP/CPP},ml}` of beamlet :math:`ml` induced by the phase plate,
    and the phase contribution to beamlet :math:`ml`, :math:`\psi_{{\rm SSD},ml}(t)`, from the temporal smoothing.
    The beam widths are :math:`\Delta x=\frac{\lambda_0fN_{bx}}{D_{x}}`,
    :math:`\Delta y=\frac{\lambda_0fN_{by}}{D_{y}}`.
    The other parameters in these formulas are defined below.

    This profile admits several options for calculating the amplitudes and phases of the beamlets:

    * Random phase plates (RPP), ``'RPP'``:
        Here the phase plate contribution to the phase of beamlet :math:`ml` is :math:`\phi_{{\rm RPP},ml}\in\{0,\pi\}` (drawn randomly for each :math:`m,l`), :math:`\psi_{{\rm SSD},ml}(t)=0`, and :math:`A_{ml}=1`
    * Continuous phase plates (CPP), ``'CPP'``:
        :math:`\phi_{{\rm CPP},ml}\in[0,\pi]` (drawn randomly with uniform probability between :math:`0` and :math:`\pi`, for each :math:`m,l`), :math:`\psi_{{\rm SSD},ml}(t)=0`, and :math:`A_{ml}=1`
    * CPP + Smoothing by spectral dispersion (SSD), ``'FM SSD'``:
        :math:`\phi_{{\rm CPP},ml}\in[0,\pi]` (drawn randomly with uniform probability between :math:`0` and :math:`\pi`, for each :math:`m,l`),

        .. math::

            \begin{aligned}
            \psi_{{\rm SSD},ml}(t)&=\delta_{x} \sin\left(\omega_{x} t + 2\pi\frac{mN_{cc,x}}{N_{bx}}\right)\\
            &+\delta_{y} \sin\left(\omega_{y} t + 2\pi\frac{lN_{cc,y}}{N_{by}}\right),
            \end{aligned}

        and :math:`A_{ml}=1`.  The modulation frequencies :math:`\omega_x,\omega_y` are determined by the
        laser bandwidth and modulation amplitudes by the relation

        .. math::

            \omega_x = \frac{\Delta_\nu r }{2\delta_x}

        where :math:`\Delta_\nu` is the relative bandwidth of the laser pulse and :math:`r` is an additional rotation factor
        supplied by the user that determines how much of the modulation is in x and how much is in y. [Michel, Eqn. 9.69]

    * Gaussian Process Randomly phase-modulated SSD, ``'GP RPM SSD'``:
        CPP + a generalization of SSD that has temporal stochastic variation in the beamlet phases; that is, :math:`\phi_{{\rm CPP},ml}\in[0,\pi]`, :math:`\psi_{{\rm SSD},ml}(t)` is sampled from a Gaussian stochastic process, and :math:`A_{ml}=1`
    * Induced spatial incoherence (ISI), ``'GP ISI'``:
        a smoothing technique with temporal stochastic variation in the beamlet phases and amplitudes; that is, :math:`\phi_{{\rm CPP},ml}=0`, :math:`\psi_{{\rm SSD},ml}(t)=0`, and the :math:`A_{ml}` are complex amplitudes sampled from a Gaussian stochastic process to simulate the random phase differences and amplitudes the beamlets pick up when passing through the ISI echelons

    This is an adapation of work by `Han Wen <https://github.com/Wen-Han/LasersSmoothing2d>`__ to LASY.


    Notes
    -----
    This assumes a flat-top rectangular laser and so a rectangular arrangement of beamlets in the near-field.

    Parameters
    ----------
    wavelength : float (in meters)
        The main laser wavelength :math:`\lambda_0` of the laser, which
        defines :math:`\omega_0` in the above formula, according to
        :math:`\omega_0 = 2\pi c/\lambda_0`.

    pol : list of 2 complex numbers (dimensionless)
        Polarization vector. It corresponds to :math:`p_u` in the above
        formula ; :math:`p_x` is the first element of the list and
        :math:`p_y` is the second element of the list. Using complex
        numbers enables elliptical polarizations.

    laser_energy : float (in Joules)
        The total energy of the laser pulse. The amplitude of the laser
        field (:math:`E_0` in the above formula) is automatically
        calculated so that the pulse has the prescribed energy.

    focal_length : float (in meters)
        Focal length of lens :math:`f` just after the RPP/CPP.

    beam_aperture : list of 2 floats (in meters)
        Widths :math:`D_x,D_y` of the rectangular beam in the near-field, i.e., size of the illuminated region of the RPP/CPP.

    n_beamlets : list of 2 integers
        Number of RPP/CPP elements :math:`N_{bx},N_{by}` in each direction, in the near field.

    temporal_smoothing_type : string
        Which method for beamlet production and evolution is used.
        Can be ``'RPP'``, ``'CPP'``, ``'FM SSD'``, ``'GP RPM SSD'``, or ``'GP ISI'``.
        (See the above bullet list for an explanation of each of these methods.)

    relative_laser_bandwidth : float
        Bandwidth :math:`\Delta_\nu` of the laser pulse, relative to central frequency.
        Only used if ``temporal_smoothing_type`` is ``'FM SSD'``, ``'GP RPM SSD'`` or ``'GP ISI'``.

    ssd_phase_modulation_amplitude :list of 2 floats
        Amplitudes :math:`\delta_{x},\delta_{y}` of phase modulation in each transverse direction.
        Only used if ``temporal_smoothing_type`` is ``'FM SSD'`` or ``'GP RPM SSD'``.

    ssd_number_color_cycles : list of 2 floats
        Number of color cycles :math:`N_{cc,x},N_{cc,y}` of SSD spectrum to include in modulation
        Only used if ``temporal_smoothing_type`` is ``'FM SSD'`` or ``'GP RPM SSD'``.

    ssd_transverse_bandwidth_distribution: list of 2 floats
        Determines how much SSD is distributed in the :math:`x` and :math:`y` directions.
        if `ssd_transverse_bandwidth_distribution=[a,b]`, then the SSD frequency modulation is :math:`a/\sqrt{a^2+b^2}` in :math:`x` and :math:`b/\sqrt{a^2+b^2}` in :math:`y`.
        Only used if ``temporal_smoothing_type`` is ``'FM SSD'`` or ``'GP RPM SSD'``.

    do_include_transverse_envelope : boolean, (optional, default False)
        Whether to include the transverse sinc envelope or not.
        I.e. whether it is assumed to be close enough to the laser axis to neglect the transverse field decay.
    """

    supported_smoothing = "RPP", "CPP", "FM SSD", "GP RPM SSD", "GP ISI"

    def __init__(
        self,
        wavelength,
        pol,
        laser_energy,
        focal_length,
        beam_aperture,
        n_beamlets,
        temporal_smoothing_type,
        relative_laser_bandwidth,
        ssd_phase_modulation_amplitude=None,
        ssd_number_color_cycles=None,
        ssd_transverse_bandwidth_distribution=None,
        do_include_transverse_envelope=False,
    ):
        super().__init__(wavelength, pol)
        self.laser_energy = laser_energy
        self.focal_length = focal_length
        self.beam_aperture = np.array(beam_aperture, dtype="float")
        self.n_beamlets = np.array(n_beamlets, dtype="int")
        self.temporal_smoothing_type = temporal_smoothing_type
        self.laser_bandwidth = relative_laser_bandwidth

        # time interval to update the speckle pattern, roughly update 50 times every bandwidth cycle
        self.dt_update = 1 / self.laser_bandwidth / 50
        self.do_include_transverse_envelope = do_include_transverse_envelope

        self.x_lens_list = np.linspace(
            -0.5 * (self.n_beamlets[0] - 1),
            0.5 * (self.n_beamlets[0] - 1),
            num=self.n_beamlets[0],
        )
        self.y_lens_list = np.linspace(
            -0.5 * (self.n_beamlets[1] - 1),
            0.5 * (self.n_beamlets[1] - 1),
            num=self.n_beamlets[1],
        )
        self.Y_lens_matrix, self.X_lens_matrix = np.meshgrid(
            self.y_lens_list, self.x_lens_list
        )
        self.Y_lens_index_matrix, self.X_lens_index_matrix = np.meshgrid(
            np.arange(self.n_beamlets[1], dtype=float),
            np.arange(self.n_beamlets[0], dtype=float),
        )

        if "SSD" in self.temporal_smoothing_type.upper():
            # Initialize SSD parameters
            # the amplitude of phase along each direction
            self.ssd_phase_modulation_amplitude = ssd_phase_modulation_amplitude
            # number of color cycles
            self.ssd_number_color_cycles = ssd_number_color_cycles
            # bandwidth distributed with respect to the two transverse direction
            self.ssd_transverse_bandwidth_distribution = (
                ssd_transverse_bandwidth_distribution
            )
            ssd_normalization = np.sqrt(
                self.ssd_transverse_bandwidth_distribution[0] ** 2
                + self.ssd_transverse_bandwidth_distribution[1] ** 2
            )
            ssd_frac = [
                self.ssd_transverse_bandwidth_distribution[0] / ssd_normalization,
                self.ssd_transverse_bandwidth_distribution[1] / ssd_normalization,
            ]
            self.ssd_phase_modulation_frequency = [
                self.laser_bandwidth * sf * 0.5 / pma
                for sf, pma in zip(ssd_frac, self.ssd_phase_modulation_amplitude)
            ]
            self.ssd_time_delay = (
                (
                    self.ssd_number_color_cycles[0]
                    / self.ssd_phase_modulation_frequency[0]
                    if self.ssd_phase_modulation_frequency[0] > 0
                    else 0
                ),
                (
                    self.ssd_number_color_cycles[1]
                    / self.ssd_phase_modulation_frequency[1]
                    if self.ssd_phase_modulation_frequency[1] > 0
                    else 0
                ),
            )

        # Check user input
        assert temporal_smoothing_type.upper() in SpeckleProfile.supported_smoothing, (
            "Only support one of the following: "
            + ", ".join(SpeckleProfile.supported_smoothing)
        )
        assert relative_laser_bandwidth > 0, "laser_bandwidth must be greater than 0"
        assert np.size(n_beamlets) == 2, "has to be a size 2 array"
        if "SSD" in self.temporal_smoothing_type.upper():
            assert ssd_number_color_cycles is not None, (
                "must supply `ssd_number_color_cycles` to use SSD"
            )
            assert ssd_transverse_bandwidth_distribution is not None, (
                "must supply `ssd_transverse_bandwidth_distribution` to use SSD"
            )
            assert ssd_phase_modulation_amplitude is not None, (
                "must supply `ssd_phase_modulation_amplitude` to use SSD"
            )
            for q in (
                ssd_number_color_cycles,
                ssd_transverse_bandwidth_distribution,
                ssd_phase_modulation_amplitude,
            ):
                assert np.size(q) == 2, "has to be a size 2 array"
                assert q[0] > 0 or q[1] > 0, "cannot be all zeros"

    def init_gaussian_time_series(
        self,
        series_time,
    ):
        r"""Initialize a time series sampled from a Gaussian process.

        At every time specified by the input `series_time`, calculate the random phase and/or amplitudes as determined by the smoothing type.

        * If the smoothing type is "SSD", then this function returns a time series with random phase offsets in x and y at each time.
            The phase offsets are real-valued and centered around the user supplied ``ssd_phase_modulation_amplitude``
            :math:`\delta_{x},\delta_{y}`, with distribution FWHM ``ssd_phase_modulation_frequency``.

        If the smoothing type is "ISI", this function returns a time series with complex numbers defining beamlet phase and amplitude.
            Each beamlet has a complex number whose magnitude is the amplitude of the beamlet and
            whose phase gives the beamlet phase offset.
            The complex numbers have mean 1 and FWHM given by twice the laser bandwidth

        Parameters
        ----------
        series_time : 1d array
            Array of times at which to sample from Gaussian process

        ssd_time_delay : scalar
            Only required for "SSD" type smoothing

        ssd_phase_modulation_frequency : scalar
            Only required for "SSD" type smoothing

        Returns
        -------
        array-like, either the supplied `series_time` if "ISI" smoothing or `series_time` with some padding at the end for "SSD" smoothing
        array-like, either with 2 (for "SSD" smoothing) or `n_beamlets[0] x n_beamlets[1]` ("ISI" smoothing) random numbers at every time
        """
        if "SSD" in self.temporal_smoothing_type.upper():
            pm_phase0 = gen_gaussian_time_series(
                series_time.size
                + int(np.sum(self.ssd_time_delay) / self.dt_update)
                + 2,
                self.dt_update,
                2 * np.pi * self.ssd_phase_modulation_frequency[0],
                self.ssd_phase_modulation_amplitude[0],
            )
            pm_phase1 = gen_gaussian_time_series(
                series_time.size
                + int(np.sum(self.ssd_time_delay) / self.dt_update)
                + 2,
                self.dt_update,
                2 * np.pi * self.ssd_phase_modulation_frequency[1],
                self.ssd_phase_modulation_amplitude[1],
            )
            time_interp = np.arange(
                start=0,
                stop=series_time[-1] + np.sum(self.ssd_time_delay) + 3 * self.dt_update,
                step=self.dt_update,
            )[: pm_phase0.size]
            return (
                time_interp,
                [
                    (np.real(pm_phase0) + np.imag(pm_phase0)) / np.sqrt(2),
                    (np.real(pm_phase1) + np.imag(pm_phase1)) / np.sqrt(2),
                ],
            )
        elif "ISI" in self.temporal_smoothing_type.upper():
            complex_amp = np.stack(
                [
                    np.stack(
                        [
                            gen_gaussian_time_series(
                                series_time.size,
                                self.dt_update,
                                2 * self.laser_bandwidth,
                                1,
                            )
                            for _i in range(self.n_beamlets[1])
                        ]
                    )
                    for _j in range(self.n_beamlets[0])
                ]
            )
            return series_time, complex_amp
        else:
            return series_time, None

    def beamlets_complex_amplitude(
        self, t_now, series_time, time_series, temporal_smoothing_type="FM SSD"
    ):
        """Calculate complex amplitude of the beamlets in the near-field, before propagating to the focal plane.

        If the temporal smoothing type is "RPP" or "CPP", this returns a matrix of ones, giving no modification to the amplitude.
        If the temporal smoothing type is "FM SSD", this returns the complex phases as calculated in, for example, Introduction to Laser-Plasma Interactions eqn. 9.87.
        If the temporal smoothing type is "GP RPM FM ", this returns complex phases modeled as random variables.
        If the temporal smoothing type is "ISI", this returns an array of random complex numbers that gives both amplitude and phase of the beamlets.

        Parameters
        ----------
        t_now : scalar
            Time at which to calculate the complex amplitude of the beamlets

        series_time : 1d array
            Array of times at which the stochastic process was sampled to generate the time series

        time_series : 1d array
            Array of random phase and/or amplitudes as determined by the smoothing type

        temporal_smoothing_type : string
            What type of temporal smoothing to perform.

        Returns
        -------
        array of complex numbers giving beamlet amplitude and phases in the near-field
        """
        if any(
            rpp_type in temporal_smoothing_type.upper() for rpp_type in ["RPP", "CPP"]
        ):
            return np.ones_like(self.X_lens_matrix)
        if temporal_smoothing_type.upper() == "FM SSD":
            phase_t = self.ssd_phase_modulation_amplitude[0] * np.sin(
                self.ssd_x_y_dephasing[0]
                + 2
                * np.pi
                * self.ssd_phase_modulation_frequency[0]
                * (
                    t_now
                    - self.X_lens_matrix * self.ssd_time_delay[0] / self.n_beamlets[0]
                )
            ) + self.ssd_phase_modulation_amplitude[1] * np.sin(
                self.ssd_x_y_dephasing[1]
                + 2
                * np.pi
                * self.ssd_phase_modulation_frequency[1]
                * (
                    t_now
                    - self.Y_lens_matrix * self.ssd_time_delay[1] / self.n_beamlets[1]
                )
            )
            return np.exp(1j * phase_t)
        elif temporal_smoothing_type.upper() == "GP RPM SSD":
            phase_t = np.interp(
                t_now
                + self.X_lens_index_matrix
                * self.ssd_time_delay[0]
                / self.n_beamlets[0],
                series_time,
                time_series[0],
            ) + np.interp(
                t_now
                + self.Y_lens_index_matrix
                * self.ssd_time_delay[1]
                / self.n_beamlets[1],
                series_time,
                time_series[1],
            )
            return np.exp(1j * phase_t)
        elif temporal_smoothing_type.upper() == "GP ISI":
            return time_series[:, :, int(round(t_now / self.dt_update))]
        else:
            raise NotImplementedError

    def generate_speckle_pattern(
        self, t_now, exp_phase_plate, x, y, series_time, time_series
    ):
        """Calculate the speckle pattern in the focal plane.

        Calculates the complex envelope defining the laser pulse in the focal plane at time `t=t_now`.
        This function first gets the beamlet complex amplitudes and phases with the function `beamlets_complex_amplitude`
        then propagates the the beamlets to the focal plane.

        Parameters
        ----------
        t_now : Scalar
            Time at which to calculate the speckle pattern

        exp_phase_plate : 2d array of complex numbers
            The RPP / CPP phase contributions to the beamlets

        x : 3d array
            x-positions in focal plane

        y : 3d array
            y-positions in focal plane

        series_time : 1d array
            Times at which the stochastic process was sampled to generate the time series

        time_series : array
            Random phase and/or amplitudes as determined by the smoothing type

        Returns
        -------
        speckle_amp : 2D array of complex numbers defining the laser envelope at focus at time `t_now`
        """
        lambda_fnum = self.lambda0 * self.focal_length / self.beam_aperture
        X_focus_matrix = x[:, :, 0] / lambda_fnum[0]
        Y_focus_matrix = y[:, :, 0] / lambda_fnum[1]
        x_focus_list = X_focus_matrix[:, 0]
        y_focus_list = Y_focus_matrix[0, :]
        x_phase_focus_matrix = np.exp(
            -2
            * np.pi
            * 1j
            / self.n_beamlets[0]
            * self.x_lens_list[:, np.newaxis]
            * x_focus_list[np.newaxis, :]
        )
        y_phase_focus_matrix = np.exp(
            -2
            * np.pi
            * 1j
            / self.n_beamlets[1]
            * self.y_lens_list[:, np.newaxis]
            * y_focus_list[np.newaxis, :]
        )

        bca = self.beamlets_complex_amplitude(
            t_now,
            series_time=series_time,
            time_series=time_series,
            temporal_smoothing_type=self.temporal_smoothing_type,
        )
        speckle_amp = np.einsum(
            "jk,jl->kl",
            np.einsum("ij,ik->jk", bca * exp_phase_plate, x_phase_focus_matrix),
            y_phase_focus_matrix,
        )
        if self.do_include_transverse_envelope:
            speckle_amp = (
                np.sinc(X_focus_matrix / self.n_beamlets[0])
                * np.sinc(Y_focus_matrix / self.n_beamlets[1])
                * speckle_amp
            )
        return speckle_amp

    def evaluate(self, x, y, t):
        """
        Return the envelope field of the laser.

        Parameters
        ----------
        x, y, t : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y, t
        """
        # General parameters
        t_norm = t[0, 0, :] * c / self.lambda0
        t_max = t_norm[-1]

        # Calculate auxiliary parameters
        if "RPP" == self.temporal_smoothing_type.upper():
            phase_plate = np.random.choice([0, np.pi], self.n_beamlets)
        elif any(
            cpp_smoothing_type in self.temporal_smoothing_type.upper()
            for cpp_smoothing_type in ["CPP", "SSD"]
        ):
            phase_plate = np.random.uniform(
                -np.pi, np.pi, size=self.n_beamlets[0] * self.n_beamlets[1]
            ).reshape(self.n_beamlets)
        elif "ISI" in self.temporal_smoothing_type.upper():
            phase_plate = np.zeros(self.n_beamlets)  # ISI does not require phase plates
        else:
            raise NotImplementedError
        exp_phase_plate = np.exp(1j * phase_plate)
        if self.temporal_smoothing_type.upper() == "FM SSD":
            self.ssd_x_y_dephasing = np.random.standard_normal(2) * np.pi

        series_time = np.arange(0, t_max + self.dt_update, self.dt_update)

        if "GP" in self.temporal_smoothing_type.upper():
            new_series_time, time_series = self.init_gaussian_time_series(series_time)
        else:
            new_series_time, time_series = series_time, None

        envelope = np.zeros(x.shape, dtype=complex)
        for i, t_i in enumerate(t_norm):
            envelope[:, :, i] = self.generate_speckle_pattern(
                t_i,
                exp_phase_plate=exp_phase_plate,
                x=x,
                y=y,
                series_time=new_series_time,
                time_series=time_series,
            )
        return envelope



import numpy as np
import openpmd_api as io
from scipy.constants import c

from lasy.utils.laser_utils import (
    create_grid,
    field_to_envelope,
    vector_potential_to_field,
)
from lasy.utils.openpmd_helper import convert_modes, extract_array, isolate_polarization

from .from_array_profile import FromArrayProfile


class FromOpenPMDProfile(FromArrayProfile):
    r"""
    Profile defined from an openPMD file.

    Upon initialization, read from an openPMD profile, build interpolation objects on the array data and use them to create function evaluate.

    Parameters
    ----------
    file_name : string
        Name of openPMD file, including path, to read the laser field or envelope from.

    envelope_name : string (optional)
        The name of the envelope field (this is not prescribed by the openPMD standard for the envelope).
        If specified, an envelope field is expected from the openPMD file. Otherwise, a full electric field is assumed.
        In the case of a full field, The transverse electric field (Ex & Ey or Er and Etheta) is read, and the polarization is measured from this.

    verbose : bool (optional)
        If true, print some intermediate steps.
    """

    def __init__(self, file_name, envelope_name=None, verbose=False):
        series = io.Series(file_name, io.Access.read_only)
        iterations = np.array(series.iterations)
        i = series.iterations[iterations[-1]]
        is_envelope = envelope_name is not None
        if is_envelope:
            if verbose:
                print("Read envelope")
            m = i.meshes[envelope_name]
            geometry = m.get_attribute("geometry")
            dim = "xyt" if geometry == "cartesian" else "rt"
            omg0 = m.get_attribute("angularFrequency")
            try:
                envelopeField = m.get_attribute("envelopeField")
                pol = m.get_attribute("polarization")
            except Exception:
                envelopeField = "normalized_vector_potential"
                pol = (1, 0)
                print(
                    "WARNING: 'envelopeField' and/or 'polarization' attributes must be specified according to the standard but are currently missing for mesh record "
                    + envelope_name
                    + ", see https://github.com/openPMD/openPMD-standard/blob/upcoming-2.0.0/EXT_LaserEnvelope.md. Assumed 'normalized_vector_potential' and (1,0), respectively."
                )
            axes_order, axes, array = extract_array(m, series)
            arrays = convert_modes([array], geometry, is_envelope, verbose)
            array = arrays[0]
            if envelopeField == "normalized_vector_potential":
                if verbose:
                    print("Convert from vector potential to electric field")
                grid = create_grid(array, axes, dim)
                array = vector_potential_to_field(grid, omg0)
        else:
            geometry = i.meshes["E"].get_attribute("geometry")
            if geometry == "cartesian":
                field_list = ["E", "E"]
                coord_list = ["x", "y"]
            else:  # thetaMode
                field_list = ["E", "E"]
                coord_list = ["r", "t"]
            array_list = []
            for count, field in enumerate(field_list):
                # Read the data
                m = i.meshes[field]
                component = coord_list[count]
                axes_order, axes, array = extract_array(m, series, component)
                array_list.append(array)
            # Convert from Er & Etheta at openPMD mode decomposition
            #           to Ex & Ey at LASY mode decomposition.
            array_list = convert_modes(array_list, geometry, is_envelope, verbose)
            dim = "xyt" if geometry == "cartesian" else "rt"
            env_array_list = []
            for array in array_list:
                grid = create_grid(array, axes, dim, is_envelope=False)
                omg0 = field_to_envelope(grid, dim)
                array = grid.get_temporal_field()
                env_array_list.append(array)
            array, pol = isolate_polarization(env_array_list, dim)
        wavelength = 2 * np.pi * c / omg0

        super().__init__(
            wavelength=wavelength,
            pol=pol,
            array=array,
            dim=dim,
            axes=axes,
            axes_order=axes_order,
        )



import numpy as np
import scipy.special as scispe

from .transverse_profile import TransverseProfile


class JincTransverseProfile(TransverseProfile):
    r"""
    Class for the analytic profile of a Jinc laser pulse.

    The transverse envelope corresponds to:

    .. math::

        \mathcal{T}(x, y) = 2\frac{J_1(r/w_0)}{r/w_0} \textrm{, with } r=\sqrt{x^2+y^2}

    where :math:`J_1` is the Bessel function of the first kind of order one

    Parameters
    ----------
    w0 : float (in meter)
        The waist of the laser pulse, i.e. :math:`w_0` in the above formula.
    """

    def __init__(self, w0):
        super().__init__()
        self.w0 = w0

    def _evaluate(self, x, y):
        """
        Return the transverse envelope.

        Parameters
        ----------
        x, y : ndarrays of floats
            Define points on which to evaluate the envelope
            These arrays need to all have the same shape.

        Returns
        -------
        envelope : ndarray of complex numbers
            Contains the value of the envelope at the specified points
            This array has the same shape as the arrays x, y
        """
        r_over_w0 = np.sqrt(x**2 + y**2) / self.w0

        envelope = np.ones_like(r_over_w0)
        # Avoid dividing by zero
        np.divide(
            2.0 * scispe.jv(1, r_over_w0),
            r_over_w0,
            out=envelope,
            where=r_over_w0 > 0.0,
        )

        return envelope



"""
Test the implementation of continuous wave laser and plane wave laser.

This test file verifies the correct implementation of these cases of the laser object and
checks the implementation of the peak_fluence and peak_power normalizations
as well as some measurement functionality from laser utils
"""

import numpy as np
from scipy.constants import c, epsilon_0

from lasy.laser import Laser
from lasy.profiles import CombinedLongitudinalTransverseProfile
from lasy.profiles.longitudinal import (
    ContinuousWaveProfile,
    GaussianLongitudinalProfile,
)
from lasy.profiles.transverse import GaussianTransverseProfile, PlaneWaveProfile
from lasy.utils.laser_utils import (
    get_duration,
    get_laser_power,
    get_w0,
)


def test_continuous_wave_laser():
    # Physical Parameters
    peak_fluence = 1e4  # J/m^2
    spot_size = 10e-3
    wavelength = 800e-9
    pol = (1, 0)

    long_prof = ContinuousWaveProfile(wavelength)
    tran_prof = GaussianTransverseProfile(spot_size)
    profile = CombinedLongitudinalTransverseProfile(
        wavelength, pol, long_prof, tran_prof, peak_fluence=peak_fluence
    )

    # Computational Grid
    dim = "xyt"
    lo = (-5 * spot_size, -5 * spot_size, None)
    hi = (5 * spot_size, 5 * spot_size, None)
    npoints = (1000, 1000, 1)

    laser = Laser(dim, lo, hi, npoints, profile)

    field = laser.grid.get_temporal_field()
    intensity = np.abs(epsilon_0 * field**2 / 2 * c)
    fluence = np.sum(intensity, axis=-1) * laser.grid.dx[-1]
    measured_peak_fluence = fluence.max()

    assert np.abs(measured_peak_fluence - peak_fluence) / peak_fluence < 1e-6
    assert np.abs(get_w0(laser.grid, laser.dim) - spot_size) / spot_size < 1e-6


def test_plane_wave_laser():
    # Physical Parameters
    tau = 30e-15
    wavelength = 800e-9
    t_peak = 0.0
    pol = (1, 0)
    peak_power = 1e12  # W

    long_prof = GaussianLongitudinalProfile(wavelength, tau, t_peak)
    tran_prof = PlaneWaveProfile()
    profile = CombinedLongitudinalTransverseProfile(
        wavelength, pol, long_prof, tran_prof, peak_power=peak_power
    )

    # Computational Grid
    dim = "xyt"
    lo = (None, None, -5 * tau)
    hi = (None, None, 5 * tau)
    npoints = (1, 1, 1000)

    laser = Laser(dim, lo, hi, npoints, profile)

    power = get_laser_power(laser.dim, laser.grid)
    measured_peak_power = power.max()

    assert np.abs(measured_peak_power - peak_power) / peak_power < 1e-6
    assert np.abs(2 * get_duration(laser.grid, laser.dim) - tau) / tau < 1e-6



from abc import ABC, abstractmethod

import numpy as np


class OpticalElement(ABC):
    """
    Base class to model thin optical elements.

    Any optical element should inherit from this class, and define its own
    `amplitude_multiplier` method, using the same signature as the method below.
    """

    def __init__(self):
        pass

    @abstractmethod
    def amplitude_multiplier(self, x, y, omega):
        r"""
        Return the amplitude multiplier :math:`T`.

        This number multiplies the complex amplitude of the laser
        just before this thin element, in order to obtain the complex
        amplitude output laser just after this thin element:

        .. math::

            \tilde{\mathcal{E}}_{out}(x, y, \omega) = T(x, y, \omega)\tilde{\mathcal{E}}_{in}(x, y, \omega)

        Parameters
        ----------
        Return the amplitude multiplier.

        Parameters
        ----------
        x, y, omega : ndarrays of floats
            Define points on which to evaluate the multiplier.
            These arrays need to all have the same shape.

        Returns
        -------
        multiplier : ndarray of complex numbers
            Contains the value of the multiplier at the specified points.
            This array has the same shape as the array omega.
        """
        # The base class only defines dummy multiplier
        # (This should be replaced by any class that inherits from this one.)
        return np.ones_like(x, dtype="complex128")



from setuptools import find_packages, setup

import lasy  # In order to extract the version number

# README file is long description
with open("./README.md", encoding="utf-8") as f:
    long_description = f.read()

# Get the package requirements from the requirements.txt files
install_requires = []
with open("./requirements.txt") as f:
    install_requires = [line.strip("\n") for line in f.readlines()]
tests_require = []
with open("./tests/requirements.txt") as f:
    tests_require = [line.strip("\n") for line in f.readlines()]

setup(
    name="lasy",
    version=lasy.__version__,
    packages=find_packages("."),
    description="LAser pulse manipulation made eaSY",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Rob Shalloo, Maxence Thevenet, Remi Lehe, Axel Huebl, Igor Adriyash, Soeren Jalas, Angel Ferran Pousa, Luca Fedeli",
    # author_email='',
    maintainer="Maxence Thevenet, Remi Lehe, Axel Huebl",
    maintainer_email="maxence.thevenet@desy.de, rlehe@lbl.gov, axelhuebl@lbl.gov",
    python_requires=">=3.7",
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        "tests": tests_require,
    },
    url="https://github.com/LASY-org/lasy",
    project_urls={
        "Documentation": "https://lasydoc.readthedocs.io",
        "Source": "https://github.com/LASY-org/lasy",
        "Tracker": "https://github.com/LASY-org/lasy/issues",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
    ],
    license="BSD-3-Clause-LBNL",
    license_files=["license.txt", "legal.txt"],
)



import numpy as np

from .optical_element import OpticalElement


class SpectralPhase(OpticalElement):
    r"""
    Class for an optical element that adds a spectral phase to a laser pulse.

    The amplitude multiplier corresponds to:

    .. math::

        T(\omega) = \exp(i(\phi(\omega)))

    where :math:`\phi(\omega)` is the spectral phase.

    Parameters
    ----------
    phase : 1D ndarray of floats (in rad)
        Phase that should be applied to the laser pulse. The phase is assumed to be unwrapped.
    omega : 1D ndarray of floats (in rad/s)
        Angular frequencies at which the phase is defined.
    """

    def __init__(self, phase, omega):
        self.phase = phase
        self.omega_in = omega

    def amplitude_multiplier(self, x, y, omega):
        """
        Return the amplitude multiplier.

        Parameters
        ----------
        x, y, omega : ndarrays of floats
            Define points on which to evaluate the multiplier.
            These arrays need to all have the same shape.

        Returns
        -------
        multiplier : ndarray of complex numbers
            Contains the value of the multiplier at the specified points.
        """
        # Sort the transmission function and angular frequencies to
        # eliminate problems with interpolation (e.g. due to fftshifted data)
        order = np.argsort(self.omega_in)
        self.omega_in = self.omega_in[order]
        self.phase = self.phase[order]

        # interpolate transmission function to omega axis of the laser
        phase = np.interp(omega, self.omega_in, self.phase)

        return np.exp(1j * phase)



from .axicon import Axicon
from .axiparabola import Axiparabola
from .intensity_mask import IntensityMask
from .parabolic_mirror import ParabolicMirror
from .polynomial_spectral_phase import PolynomialSpectralPhase
from .spectral_filter import SpectralFilter
from .spectral_phase import SpectralPhase
from .zernike_aberrations import ZernikeAberrations

__all__ = [
    "ParabolicMirror",
    "PolynomialSpectralPhase",
    "Axiparabola",
    "Axicon",
    "IntensityMask",
    "ZernikeAberrations",
    "SpectralFilter",
    "SpectralPhase",
]



import numpy as np
from scipy import special
from scipy.constants import c, epsilon_0

from lasy.laser import Laser
from lasy.optical_elements import Axicon
from lasy.profiles.gaussian_profile import GaussianProfile

wavelength = 0.8e-6
w0 = 10.0e-5  # m, initialized in near field
pol = (1, 0)
laser_energy = 1.0  # J
t_peak = 0.0e-15  # s
tau = 30.0e-15  # s
gaussian_profile = GaussianProfile(wavelength, pol, laser_energy, w0, tau, t_peak)
gamma = 2.2 * np.pi / 180
z_max = w0 * np.cos(gamma) / np.sin(gamma)


def check_axicon(laser):
    # Apply the axicon
    laser.apply_optics(Axicon(gamma=gamma))

    # Compare check the laser profiles
    # after different propagation distances
    distances = [0, z_max / 1.5, z_max]

    # i_dist starts at 1: we do not compare the profile at distance 0
    for i_dist in range(1, len(distances)):
        laser.propagate(distances[i_dist] - distances[i_dist - 1])
        # Check the Bessel profile
        error = check_bessel_profile(laser, distances[i_dist])
        assert error < 3e-2


def check_bessel_profile(laser, z):
    # Calculate the laser profile
    field = laser.grid.get_temporal_field()
    if laser.dim == "xyt":
        Nx = field.shape[0]
        A2 = 0.5 * epsilon_0 * c * (np.abs(field[Nx // 2 - 1, :, :]) ** 2).max(-1)
        ax = laser.grid.axes[1]
    else:
        A2 = 0.5 * epsilon_0 * c * (np.abs(field[0, :, :]) ** 2).max(-1)
        ax = laser.grid.axes[0]
        if ax[0] > 0:
            A2 = np.r_[A2[::-1], A2]
            ax = np.r_[-ax[::-1], ax]
        else:
            A2 = np.r_[A2[::-1][:-1], A2]
            ax = np.r_[-ax[::-1][:-1], ax]

    # Calculate the expected Bessel profile
    k = 2 * np.pi / wavelength
    r = np.abs(ax)
    kr = k * np.sin(gamma) * r
    power = laser_energy / (np.sqrt(np.pi / 2) * tau)

    # Equation 3 in Brzobohaty, O., Cizmar, T., & Zemanek, P. (n.d.). High quality quasi-Bessel beam generated by round-tip axicon.
    expected_profile = (
        4
        * power
        * k
        * np.sin(gamma)
        / w0
        * np.exp(-2 * z**2 / (z_max**2))
        * z
        / z_max
        * special.j0(kr) ** 2
    )

    # Calculate error
    error = np.max(np.abs(expected_profile - A2)) / abs(A2).max()
    return error


def test_3D_case():
    # 3D case
    dim = "xyt"
    lo = (-20e-5, -20e-5, -100e-15)
    hi = (+20e-5, +20e-5, +100e-15)
    npoints = (500, 500, 150)

    laser = Laser(dim, lo, hi, npoints, gaussian_profile)
    check_axicon(laser)


def test_RT_case():
    # Cylindrical case
    dim = "rt"
    lo = (0e-6, -100e-15)
    hi = (20e-5, +100e-15)
    npoints = (250, 150)

    laser = Laser(dim, lo, hi, npoints, gaussian_profile)
    check_axicon(laser)



# -*- coding: utf-8 -*-

import numpy as np
import pytest

from lasy.laser import Laser
from lasy.profiles.gaussian_profile import GaussianProfile


@pytest.fixture(scope="function")
def gaussian():
    # Cases with Gaussian laser
    wavelength = 0.8e-6
    pol = (1, 0)
    laser_energy = 1.0  # J
    t_peak = 0.0e-15  # s
    tau = 30.0e-15  # s
    w0 = 5.0e-6  # m
    profile = GaussianProfile(wavelength, pol, laser_energy, w0, tau, t_peak)

    return profile


def get_w0(laser):
    # Calculate the laser waist
    field = laser.grid.get_temporal_field()
    if laser.dim == "xyt":
        Nx, Ny, Nt = field.shape
        A2 = (np.abs(field[Nx // 2 - 1, :, :]) ** 2).sum(-1)
        ax = laser.grid.axes[1]
    else:
        A2 = (np.abs(field[0, :, :]) ** 2).sum(-1)
        ax = laser.grid.axes[0]
        if ax[0] > 0:
            A2 = np.r_[A2[::-1], A2]
            ax = np.r_[-ax[::-1], ax]
        else:
            A2 = np.r_[A2[::-1][:-1], A2]
            ax = np.r_[-ax[::-1][:-1], ax]

    sigma = 2 * np.sqrt(np.average(ax**2, weights=A2))

    return sigma


def check_gaussian_propagation(
    laser, propagation_distance=100e-6, propagation_step=10e-6
):
    # Do the propagation and check evolution of waist with theory
    w0 = laser.profile.w0
    L_R = np.pi * w0**2 / laser.profile.lambda0

    propagated_distance = 0.0
    while propagated_distance <= propagation_distance:
        propagated_distance += propagation_step
        laser.propagate(
            propagation_step,
        )
        w0_num = get_w0(laser)
        w0_theor = w0 * np.sqrt(1 + (propagated_distance / L_R) ** 2)
        err = 2 * np.abs(w0_theor - w0_num) / (w0_theor + w0_num)
        assert err < 1e-3


def test_3D_case(gaussian):
    # - 3D case
    dim = "xyt"
    lo = (-25e-6, -25e-6, -60e-15)
    hi = (+25e-6, +25e-6, +60e-15)
    npoints = (100, 102, 100)

    laser = Laser(dim, lo, hi, npoints, gaussian)
    check_gaussian_propagation(laser)


def test_RT_case(gaussian):
    # - Cylindrical case
    dim = "rt"
    lo = (0e-6, -60e-15)
    hi = (25e-6, +60e-15)
    npoints = (100, 100)

    laser = Laser(dim, lo, hi, npoints, gaussian)
    check_gaussian_propagation(laser)


def test_RT_case_multimode(gaussian):
    # - Cylindrical case
    dim = "rt"
    lo = (0e-6, -60e-15)
    hi = (25e-6, +60e-15)
    npoints = (100, 100)

    # Note: using 3 modes and 20 points is unnecessary here,
    # since the profile is purely cylindrical. This is
    # done here only to make sure that the code is robust.
    laser = Laser(dim, lo, hi, npoints, gaussian, n_azimuthal_modes=3, n_theta_evals=20)
    check_gaussian_propagation(laser)



import numpy as np
import pytest
from scipy.constants import c

np.random.seed(0)  # Fix random seed for reproducibility

from lasy.laser import Laser
from lasy.profiles.speckle_profile import SpeckleProfile


@pytest.mark.parametrize(
    "temporal_smoothing_type", ["RPP", "CPP", "FM SSD", "GP RPM SSD", "GP ISI"]
)
def test_intensity_distribution(temporal_smoothing_type):
    """Test whether the spatial intensity distribution and statisticis are correct.

    The distribution should be exponential, 1/<I> exp(-I/<I>) [Michel, 9.35].
    The real and imaginary parts of the envelope [Michel, Eqn. 9.26] and their product [9.30] should all be 0 on average.
    """
    wavelength = 0.351e-6  # Laser wavelength in meters
    polarization = (1, 0)  # Linearly polarized in the x direction
    focal_length = 3.5  # m
    beam_aperture = [0.35, 0.5]  # m
    n_beamlets = [24, 32]
    relative_laser_bandwidth = 0.005
    laser_energy = 1.0  # J (this is the laser energy stored in the box defined by `lo` and `hi` below)

    ssd_phase_modulation_amplitude = (4.1, 4.5)
    ssd_number_color_cycles = [1.4, 1.0]
    ssd_transverse_bandwidth_distribution = [1.8, 1.0]

    profile = SpeckleProfile(
        wavelength,
        polarization,
        laser_energy,
        focal_length,
        beam_aperture,
        n_beamlets,
        temporal_smoothing_type=temporal_smoothing_type,
        relative_laser_bandwidth=relative_laser_bandwidth,
        ssd_phase_modulation_amplitude=ssd_phase_modulation_amplitude,
        ssd_number_color_cycles=ssd_number_color_cycles,
        ssd_transverse_bandwidth_distribution=ssd_transverse_bandwidth_distribution,
    )
    dimensions = "xyt"
    dx = wavelength * focal_length / beam_aperture[0]
    dy = wavelength * focal_length / beam_aperture[1]
    Lx = 1.8 * dx * n_beamlets[0]
    Ly = 3.1 * dy * n_beamlets[1]
    nu_laser = c / wavelength
    t_max = 50 / nu_laser
    lo = (0, 0, 0)
    hi = (Lx, Ly, t_max)
    num_points = (200, 250, 2)

    laser = Laser(dimensions, lo, hi, num_points, profile)

    F = laser.grid.get_temporal_field()

    # get spatial statistics
    # <real env> = 0 = <imag env> = <er * ei>
    e_r = np.real(F)
    e_i = np.imag(F)
    er_ei = e_r * e_i
    assert np.max(abs(e_r.mean(axis=(0, 1)) / e_r.std(axis=(0, 1)))) < 1.0e-1
    assert np.max(abs(e_i.mean(axis=(0, 1)) / e_i.std(axis=(0, 1)))) < 1.0e-1
    assert np.max(abs(er_ei.mean(axis=(0, 1)) / er_ei.std(axis=(0, 1)))) < 1.0e-1

    # # compare intensity distribution with expected 1/<I> exp(-I/<I>)
    env_I = abs(F) ** 2
    I_vec = env_I.flatten()
    mean_I = I_vec.mean()
    N_hist = 200
    counts_np, bins_np = np.histogram(I_vec, bins=N_hist, density=True)
    I_dist = 1.0 / mean_I * np.exp(-bins_np / mean_I)
    error_I_dist = np.max(abs(counts_np - I_dist[:-1]))
    assert error_I_dist < 2.0e-4


@pytest.mark.parametrize(
    "temporal_smoothing_type", ["RPP", "CPP", "FM SSD", "GP RPM SSD", "GP ISI"]
)
def test_spatial_correlation(temporal_smoothing_type):
    """Tests whether the speckles have the correct shape.

    The speckle shape is measured over one period, since the spatial profile is periodic.
    The correct speckle shape for a rectangular laser,
    determined by the autocorrelation, is the product of sinc functions [Michel, Eqn. 9.16].
    """
    wavelength = 0.351e-6  # Laser wavelength in meters
    polarization = (1, 0)  # Linearly polarized in the x direction
    laser_energy = 1.0  # J (this is the laser energy stored in the box defined by `lo` and `hi` below)
    focal_length = 3.5  # m
    beam_aperture = [0.35, 0.35]  # m
    n_beamlets = [24, 32]
    relative_laser_bandwidth = 0.005

    ssd_phase_modulation_amplitude = (4.1, 4.1)
    ssd_number_color_cycles = [1.4, 1.0]
    ssd_transverse_bandwidth_distribution = [1.0, 1.0]

    profile = SpeckleProfile(
        wavelength,
        polarization,
        laser_energy,
        focal_length,
        beam_aperture,
        n_beamlets,
        temporal_smoothing_type=temporal_smoothing_type,
        relative_laser_bandwidth=relative_laser_bandwidth,  # 0.005
        ssd_phase_modulation_amplitude=ssd_phase_modulation_amplitude,
        ssd_number_color_cycles=ssd_number_color_cycles,
        ssd_transverse_bandwidth_distribution=ssd_transverse_bandwidth_distribution,
    )
    dimensions = "xyt"
    dx = wavelength * focal_length / beam_aperture[0]
    dy = wavelength * focal_length / beam_aperture[1]
    Lx = dx * n_beamlets[0]
    Ly = dy * n_beamlets[1]
    nu_laser = c / wavelength
    tu = 1 / relative_laser_bandwidth / 50 / nu_laser
    t_max = 200 * tu
    lo = (0, 0, 0)
    hi = (Lx, Ly, t_max)
    num_points = (200, 200, 300)

    laser = Laser(dimensions, lo, hi, num_points, profile)
    F = laser.grid.get_temporal_field()

    # compare speckle profile / autocorrelation
    # compute autocorrelation using Wiener-Khinchin Theorem

    fft_abs_all = abs(np.fft.fft2(F, axes=(0, 1))) ** 2
    ifft_abs = abs(np.fft.ifft2(fft_abs_all, axes=(0, 1))) ** 2
    acorr2_3d = np.fft.fftshift(ifft_abs, axes=(0, 1))
    acorr2_3d_norm = acorr2_3d / np.max(acorr2_3d, axis=(0, 1))

    # compare with theoretical speckle profile
    x_list = np.linspace(
        -n_beamlets[0] / 2 + 0.5, n_beamlets[0] / 2 - 0.5, num_points[0], endpoint=False
    )
    y_list = np.linspace(
        -n_beamlets[1] / 2 + 0.5, n_beamlets[1] / 2 - 0.5, num_points[1], endpoint=False
    )
    X, Y = np.meshgrid(x_list, y_list)
    acorr_theor = np.sinc(X) ** 2 * np.sinc(Y) ** 2
    error_auto_correlation = np.max(abs(acorr_theor[:, :, np.newaxis] - acorr2_3d_norm))

    assert error_auto_correlation < 5.0e-1


@pytest.mark.parametrize(
    "temporal_smoothing_type", ["RPP", "CPP", "FM SSD", "GP RPM SSD", "GP ISI"]
)
def test_sinc_zeros(temporal_smoothing_type):
    r"""Test whether the transverse sinc envelope has the correct width.

    The transverse envelope for the rectangular laser has the form

    ..math::

        {\rm sinc}\left(\frac{\pi x}{\Delta x}\right)
        {\rm sinc}\left(\frac{\pi y}{\Delta y}\right)

    [Michel, Eqns. 9.11, 87, 94].
    This has widths

    ..math::

        \Delta x=\lambda_0fN_{bx}/D_x,
        \Delta y=\lambda_0fN_{by}/D_y
    """
    wavelength = 0.351e-6  # Laser wavelength in meters
    polarization = (1, 0)  # Linearly polarized in the x direction
    laser_energy = 1.0  # J (this is the laser energy stored in the box defined by `lo` and `hi` below)
    focal_length = 3.5  # m
    beam_aperture = [0.35, 0.35]  # m
    n_beamlets = [24, 48]
    relative_laser_bandwidth = 0.005
    ssd_phase_modulation_amplitude = (4.1, 4.1)
    ssd_number_color_cycles = [1.4, 1.0]
    ssd_transverse_bandwidth_distribution = [1.0, 1.0]

    profile = SpeckleProfile(
        wavelength,
        polarization,
        laser_energy,
        focal_length,
        beam_aperture,
        n_beamlets,
        temporal_smoothing_type=temporal_smoothing_type,
        relative_laser_bandwidth=relative_laser_bandwidth,
        ssd_phase_modulation_amplitude=ssd_phase_modulation_amplitude,
        ssd_number_color_cycles=ssd_number_color_cycles,
        ssd_transverse_bandwidth_distribution=ssd_transverse_bandwidth_distribution,
        do_include_transverse_envelope=True,
    )
    dimensions = "xyt"
    dx = wavelength * focal_length / beam_aperture[0]
    dy = wavelength * focal_length / beam_aperture[1]
    Lx = dx * n_beamlets[0]
    Ly = dy * n_beamlets[1]
    nu_laser = c / wavelength
    tu = 1 / relative_laser_bandwidth / 50 / nu_laser
    t_max = 200 * tu
    lo = (-Lx, -Ly, 0)
    hi = (Lx, Ly, t_max)
    num_points = (300, 300, 10)

    laser = Laser(dimensions, lo, hi, num_points, profile)
    F = laser.grid.get_temporal_field()

    assert abs(F[0, :, :]).max() / abs(F).max() < 1.0e-8
    assert abs(F[-1, :, :]).max() / abs(F).max() < 1.0e-8
    assert abs(F[:, 0, :]).max() / abs(F).max() < 1.0e-8
    assert abs(F[:, -1, :]).max() / abs(F).max() < 1.0e-8


def test_FM_SSD_periodicity():
    """Test that the frequency modulated Smoothing by spectral dispersion (SSD) has the correct temporal frequency."""
    wavelength = 0.351e-6  # Laser wavelength in meters
    polarization = (1, 0)  # Linearly polarized in the x direction
    laser_energy = 1.0  # J (this is the laser energy stored in the box defined by `lo` and `hi` below)
    focal_length = 3.5  # m
    beam_aperture = [0.35, 0.35]  # m
    n_beamlets = [24, 32]
    temporal_smoothing_type = "FM SSD"
    relative_laser_bandwidth = 0.005

    ssd_phase_modulation_amplitude = [4.1, 4.1]
    ssd_number_color_cycles = [1.4, 1.0]
    ssd_transverse_bandwidth_distribution = [1.0, 1.0]

    laser_profile = SpeckleProfile(
        wavelength,
        polarization,
        laser_energy,
        focal_length,
        beam_aperture,
        n_beamlets,
        temporal_smoothing_type=temporal_smoothing_type,
        relative_laser_bandwidth=relative_laser_bandwidth,
        ssd_phase_modulation_amplitude=ssd_phase_modulation_amplitude,
        ssd_number_color_cycles=ssd_number_color_cycles,
        ssd_transverse_bandwidth_distribution=ssd_transverse_bandwidth_distribution,
    )
    nu_laser = c / wavelength
    ssd_frac = np.sqrt(
        ssd_transverse_bandwidth_distribution[0] ** 2
        + ssd_transverse_bandwidth_distribution[1] ** 2
    )
    ssd_frac = (
        ssd_transverse_bandwidth_distribution[0] / ssd_frac,
        ssd_transverse_bandwidth_distribution[1] / ssd_frac,
    )
    phase_mod_freq = [
        relative_laser_bandwidth * sf * 0.5 / pma
        for sf, pma in zip(ssd_frac, ssd_phase_modulation_amplitude)
    ]
    t_max = 1.0 / phase_mod_freq[0] / nu_laser

    dimensions = "xyt"
    dx = wavelength * focal_length / beam_aperture[0]
    dy = wavelength * focal_length / beam_aperture[1]
    Lx = dx * n_beamlets[0]
    Ly = dy * n_beamlets[1]
    lo = (0, 0, 0)  # Lower bounds of the simulation box
    hi = (Lx, Ly, t_max)  # Upper bounds of the simulation box
    num_points = (160, 200, 400)  # Number of points in each dimension

    laser = Laser(dimensions, lo, hi, num_points, laser_profile)
    F = laser.grid.get_temporal_field()
    period_error = abs(F[:, :, 0] - F[:, :, -1]).max() / abs(F).max()
    assert period_error < 1.0e-8
